from __future__ import annotations

from time import sleep
from typing import Dict, List, Tuple

import yaml
from dateutil import parser
from oc_meta.plugins.editor import MetaEditor
from oc_ocdm.graph import GraphEntity
from oc_ocdm.graph.graph_entity import GraphEntity
from oc_ocdm.prov.prov_entity import ProvEntity
from oc_ocdm.support import get_count, get_resource_number
from rdflib import RDF, XSD, Graph, Literal, URIRef
from SPARQLWrapper import JSON, POST, SPARQLWrapper
from time_agnostic_library.agnostic_entity import AgnosticEntity


class ResourceFinder:

    def __init__(self, ts_url, base_iri:str, local_g: Graph = Graph(), settings: dict = dict(), meta_config_path: str = None):
        self.ts = SPARQLWrapper(ts_url)
        self.ts.setMethod(POST)
        self.base_iri = base_iri[:-1] if base_iri[-1] == '/' else base_iri
        self.local_g = local_g
        self.ids_in_local_g = set()
        self.meta_config_path = meta_config_path
        self.meta_settings = settings
        self.blazegraph_full_text_search = settings['blazegraph_full_text_search'] if settings and 'blazegraph_full_text_search' in settings else False
        self.virtuoso_full_text_search = settings['virtuoso_full_text_search'] if settings and 'virtuoso_full_text_search' in settings else False

    def __query(self, query, return_format = JSON):
        """Execute a SPARQL query with retries and exponential backoff"""
        self.ts.setReturnFormat(return_format)
        self.ts.setQuery(query)
        
        max_retries = 5  # Aumentiamo il numero di tentativi
        base_wait = 5    # Tempo base di attesa in secondi
        
        for attempt in range(max_retries):
            try:
                result = self.ts.queryAndConvert()
                return result
            except Exception as e:
                wait_time = base_wait * (2 ** attempt)  # Exponential backoff
                if attempt < max_retries - 1:  # Se non è l'ultimo tentativo
                    sleep(wait_time)
                else:
                    # Ultimo tentativo fallito, logghiamo l'errore e solleviamo un'eccezione custom
                    error_msg = f"Failed to execute SPARQL query after {max_retries} attempts: {str(e)}\nQuery: {query}"
                    print(error_msg)  # Log dell'errore
                    raise Exception(error_msg)

    # _______________________________BR_________________________________ #

    def retrieve_br_from_id(self, schema: str, value: str) -> List[Tuple[str, str, list]]:
        '''
        Given an identifier, it retrieves bibliographic resources associated with that identifier, related titles and other identifiers MetaIDs and literal values.

        :params schema: an identifier schema
        :type schema: str
        :params value: an identifier literal value
        :type value: str
        :returns List[Tuple[str, str, list]]: -- it returns a list of three elements tuples. The first element is the MetaID of a resource associated with the input ID. The second element is a title of that resource, if present. The third element is a list of MetaID-ID tuples related to identifiers associated with that resource. 
        '''
        schema_uri = URIRef(GraphEntity.DATACITE + schema)
        value = value.replace('\\', '\\\\')
        result_list = []
        identifier_uri = None

        # Search for both string-typed and untyped literals
        for literal_value in [Literal(value, datatype=XSD.string), Literal(value)]:
            for starting_triple in self.local_g.triples((None, GraphEntity.iri_has_literal_value, literal_value)):
                for known_id_triple in self.local_g.triples((starting_triple[0], None, None)):
                    if known_id_triple[1] == GraphEntity.iri_uses_identifier_scheme and known_id_triple[2] == schema_uri:
                        identifier_uri = known_id_triple[0]
                if identifier_uri:
                    break
        if identifier_uri:
            metaid_id_list = [(identifier_uri.replace(f'{self.base_iri}/id/', ''), f'{schema}:{value}')]
            for triple in self.local_g.triples((None, GraphEntity.iri_has_identifier, identifier_uri)):
                title = ''
                res = triple[0]
                for res_triple in self.local_g.triples((res, None, None)):
                    if res_triple[1] == GraphEntity.iri_title:
                        title = str(res_triple[2])
                    elif res_triple[1] == GraphEntity.iri_has_identifier and res_triple[2] != identifier_uri:
                        for id_triple in self.local_g.triples((res_triple[2], None, None)):
                            if id_triple[1] == GraphEntity.iri_uses_identifier_scheme:
                                id_schema = id_triple[2]
                            elif id_triple[1] == GraphEntity.iri_has_literal_value:
                                id_literal_value = id_triple[2]
                        full_id = f'{id_schema.replace(GraphEntity.DATACITE, "")}:{id_literal_value}'
                        metaid_id_tuple = (res_triple[2].replace(f'{self.base_iri}/id/', ''), full_id)
                        metaid_id_list.append(metaid_id_tuple)
                result_list.append((res.replace(f'{self.base_iri}/br/', ''), title, metaid_id_list))

        return result_list
        
    def retrieve_br_from_meta(self, metaid: str) -> Tuple[str, List[Tuple[str, str]], bool]:
        '''
        Given a MetaID, it retrieves the title of the bibliographic resource having that MetaID and other identifiers of that entity.

        :params metaid: a MetaID
        :type metaid: str
        :returns Tuple[str, List[Tuple[str, str]]]: -- it returns a tuple of two elements. The first element is the resource's title associated with the input MetaID. The second element is a list of MetaID-ID tuples related to identifiers associated with that entity.
        '''
        metaid_uri = f'{self.base_iri}/br/{metaid}'
        title = ''
        identifiers = []
        it_exists = False

        for triple in self.local_g.triples((URIRef(metaid_uri), None, None)):
            it_exists = True
            if triple[1] == GraphEntity.iri_title:
                title = str(triple[2])
            elif triple[1] == GraphEntity.iri_has_identifier:
                id_scheme = ''
                literal_value = ''
                identifier = triple[2]
                for triple_inner in self.local_g.triples((identifier, None, None)):
                    if triple_inner[1] == GraphEntity.iri_uses_identifier_scheme:
                        id_scheme = str(triple_inner[2]).replace(GraphEntity.DATACITE, '')
                    elif triple_inner[1] == GraphEntity.iri_has_literal_value:
                        literal_value = str(triple_inner[2])
                if id_scheme and literal_value:  # Ensure both id_scheme and literal_value are found before appending
                    full_id = f'{id_scheme}:{literal_value}'
                    identifiers.append((str(identifier).replace(self.base_iri + '/id/', ''), full_id))

        if not it_exists:
            return "", [], False

        return title, identifiers, True

    # _______________________________ID_________________________________ #

    def retrieve_metaid_from_id(self, schema: str, value: str) -> str:
        '''
        Given the schema and value of an ID, it returns the MetaID associated with that identifier.

        :params schema: an identifier schema
        :type schema: str
        :params value: an identifier literal value
        :type value: str
        :returns str: -- it returns the MetaID associated with the input ID.
        '''
        schema_uri = URIRef(GraphEntity.DATACITE + schema)
        value = value.replace('\\', '\\\\')
        
        # Create both untyped and string-typed literals        
        for literal in [Literal(value, datatype=XSD.string), Literal(value)]:
            for starting_triple in self.local_g.triples((None, GraphEntity.iri_has_literal_value, literal)):
                for known_id_triple in self.local_g.triples((starting_triple[0], None, None)):
                    if known_id_triple[1] == GraphEntity.iri_uses_identifier_scheme and known_id_triple[2] == schema_uri:
                        return known_id_triple[0].replace(f'{self.base_iri}/id/', '')
        
        # If no match is found, return None or an appropriate value
        return None

    def retrieve_metaid_from_merged_entity(self, metaid_uri:str, prov_config:str) -> str:
        '''
        It looks for MetaId in the provenance. If the input entity was deleted due to a merge, this function returns the target entity. Otherwise, it returns None.

        :params metaid_uri: a MetaId URI
        :type metaid_uri: str
        :params prov_config: the path of the configuration file required by time-agnostic-library
        :type prov_config: str
        :returns str: -- It returns the MetaID associated with the target entity after a merge. If there was no merge, it returns None.
        '''
        metaval = None
        with open(prov_config, 'r', encoding='utf8') as f:
            prov_config_dict = yaml.safe_load(f)
        agnostic_meta = AgnosticEntity(res=metaid_uri, config=prov_config_dict, related_entities_history=False)
        agnostic_meta_history = agnostic_meta.get_history(include_prov_metadata=True)
        meta_history_data = agnostic_meta_history[0][metaid_uri]
        if meta_history_data:
            meta_history_metadata = agnostic_meta_history[1][metaid_uri]
            penultimate_snapshot = sorted(
                meta_history_metadata.items(),
                key=lambda x: parser.parse(x[1]['generatedAtTime']).replace(tzinfo=None),
                reverse=True
            )[1][0]
            query_if_it_was_merged = f'''
                SELECT DISTINCT ?se
                WHERE {{
                    ?se a <{ProvEntity.iri_entity}>;
                        <{ProvEntity.iri_was_derived_from}> <{penultimate_snapshot}>.
                }}
            '''
            results = self.__query(query_if_it_was_merged)['results']['bindings']
            # The entity was merged to another
            merged_entity = [se for se in results if metaid_uri not in se['se']['value']]
            if merged_entity:
                merged_entity:str = merged_entity[0]['se']['value']
                merged_entity = merged_entity.split('/prov/')[0]
                merged_entity = get_count(merged_entity)
                metaval = merged_entity
            return metaval

    # _______________________________RA_________________________________ #
    def retrieve_ra_from_meta(self, metaid: str) -> Tuple[str, List[Tuple[str, str]]]:
        '''
        Given a MetaID, it retrieves the name and id of the responsible agent associated with it, whether it is an author or a publisher.
        The output has the following format:

            ('NAME', [('METAID_OF_THE_IDENTIFIER', 'LITERAL_VALUE')])
            ('American Medical Association (ama)', [('4274', 'crossref:10')])

        :params metaid: a responsible agent's MetaID
        :type metaid: str
        :returns str: -- it returns a tuple, where the first element is the responsible agent's name, and the second element is a list containing its identifier's MetaID and literal value
        '''
        metaid_uri = f'{self.base_iri}/ra/{metaid}'
        family_name = ''
        given_name = ''
        name = ''
        identifiers = []
        it_exists = False

        for triple in self.local_g.triples((URIRef(metaid_uri), None, None)):
            it_exists = True
            if triple[1] == GraphEntity.iri_family_name:
                family_name = str(triple[2])
            elif triple[1] == GraphEntity.iri_given_name:
                given_name = str(triple[2])
            elif triple[1] == GraphEntity.iri_name:
                name = str(triple[2])
            elif triple[1] == GraphEntity.iri_has_identifier:
                identifier = triple[2]
                id_scheme = ''
                literal_value = ''
                for triple_inner in self.local_g.triples((identifier, None, None)):
                    if triple_inner[1] == GraphEntity.iri_uses_identifier_scheme:
                        id_scheme = str(triple_inner[2]).replace(GraphEntity.DATACITE, '')
                    elif triple_inner[1] == GraphEntity.iri_has_literal_value:
                        literal_value = str(triple_inner[2])
                if id_scheme and literal_value:
                    full_id = f'{id_scheme}:{literal_value}'
                    identifiers.append((str(identifier).replace(self.base_iri + '/id/', ''), full_id))
        
        full_name = self._construct_full_name(name, family_name, given_name)
            
        return full_name, identifiers, it_exists

    def retrieve_ra_from_id(self, schema: str, value: str, publisher: bool) -> List[Tuple[str, str, list]]:
        '''
        Given an identifier, it retrieves responsible agents associated with that identifier, related names and other identifiers MetaIDs and literal values.
        The output has the following format: ::

            [(METAID, NAME, [(METAID_OF_THE_IDENTIFIER, LITERAL_VALUE)])]
            [('3309', 'American Medical Association (ama)', [('4274', 'crossref:10')])]

        :params schema: an identifier schema
        :type schema: str
        :params value: an identifier literal value
        :type value: str
        :params publisher: True if the identifier is associated with a publisher, False otherwise.
        :type publisher: bool
        :returns List[Tuple[str, str, list]]: -- it returns a list of three elements tuples. The first element is the MetaID of a responsible agent associated with the input ID. The second element is the name of that responsible agent, if present. The third element is a list of MetaID-ID tuples related to identifiers associated with that responsible agent. 
        '''
        schema_uri = URIRef(GraphEntity.DATACITE + schema)
        value = value.replace('\\', '\\\\')
        result_list = []
        identifier_uri = None

        # Search for both string-typed and untyped literals
        for literal_value in [Literal(value, datatype=XSD.string), Literal(value)]:
            for starting_triple in self.local_g.triples((None, GraphEntity.iri_has_literal_value, literal_value)):
                for known_id_triple in self.local_g.triples((starting_triple[0], None, None)):
                    if known_id_triple[1] == GraphEntity.iri_uses_identifier_scheme and known_id_triple[2] == schema_uri:
                        identifier_uri = known_id_triple[0]
                        break
                if identifier_uri:
                    break
        if identifier_uri:
            metaid_id_list = [(identifier_uri.replace(f'{self.base_iri}/id/', ''), f'{schema}:{value}')]
            for triple in self.local_g.triples((None, GraphEntity.iri_has_identifier, identifier_uri)):
                name = ''
                family_name = ''
                given_name = ''
                res = triple[0]
                for res_triple in self.local_g.triples((res, None, None)):
                    if res_triple[1] == GraphEntity.iri_name:
                        name = str(res_triple[2])
                    elif res_triple[1] == GraphEntity.iri_family_name:
                        family_name = str(res_triple[2])
                    elif res_triple[1] == GraphEntity.iri_given_name:
                        given_name = str(res_triple[2])
                    elif res_triple[1] == GraphEntity.iri_has_identifier and res_triple[2] != identifier_uri:
                        for id_triple in self.local_g.triples((res_triple[2], None, None)):
                            if id_triple[1] == GraphEntity.iri_uses_identifier_scheme:
                                id_schema = id_triple[2]
                            elif id_triple[1] == GraphEntity.iri_has_literal_value:
                                id_literal_value = id_triple[2]
                        full_id = f'{id_schema.replace(GraphEntity.DATACITE, "")}:{id_literal_value}'
                        metaid_id_tuple = (res_triple[2].replace(f'{self.base_iri}/id/', ''), full_id)
                        metaid_id_list.append(metaid_id_tuple)
                
                full_name = self._construct_full_name(name, family_name, given_name)
                result_list.append((res.replace(f'{self.base_iri}/ra/', ''), full_name, metaid_id_list))

        return result_list

    def _construct_full_name(self, name: str, family_name: str, given_name: str) -> str:
        if name and not family_name and not given_name:
            return name
        elif not name and family_name and not given_name:
            return f'{family_name},'
        elif not name and not family_name and given_name:
            return f', {given_name}'
        elif not name and family_name and given_name:
            return f'{family_name}, {given_name}'
        else:
            return ''

    # _______________________________VVI_________________________________ #

    def retrieve_venue_from_meta(self, meta_id:str) -> Dict[str, Dict[str, str]]:
        '''
        Given a MetaID, it returns the structure of volumes and issues contained in the related venue.
        The output has the following format: ::

            {
                'issue': {SEQUENCE_IDENTIFIER: {'id': META_ID}},
                'volume': {
                    SEQUENCE_IDENTIFIER: {
                        'id': META_ID,
                        'issue' {SEQUENCE_IDENTIFIER: {'id': META_ID}}
                    }
                }
            }

            {
                'issue': {}, 
                'volume': {
                    '166': {'id': '4388', 'issue': {'4': {'id': '4389'}}}, 
                    '172': {'id': '4434', 
                        'issue': {
                            '22': {'id': '4435'}, 
                            '20': {'id': '4436'}, 
                            '21': {'id': '4437'}, 
                            '19': {'id': '4438'}
                        }
                    }
                }
            }    

        The first level 'issue' field includes the issues contained directly in the venue, 
        while the 'volume' field includes the volumes in the venue and the related issues.

        :params meta_id: a MetaID
        :type meta_id: str
        :returns: Dict[str, Dict[str, str]] -- the string with normalized hyphens
        '''
        content = dict()
        content['issue'] = dict()
        content['volume'] = dict()
        content = self.__retrieve_vvi(meta_id, content)
        return content

    def __retrieve_vvi(self, meta:str, content:Dict[str, dict]) -> dict:
        venue_iri = URIRef(f'{self.base_iri}/br/{meta}')
        ress = []
        for triple in self.local_g.triples((None, GraphEntity.iri_part_of, venue_iri)):
            res = {'res': None, 'type': None, 'sequence_identifier': None, 'container': None}
            res['res'] = triple[0].replace(f'{self.base_iri}/br/', '')
            for res_triple in self.local_g.triples((triple[0], None, None)):
                if res_triple[1] == RDF.type and res_triple[2] != GraphEntity.iri_expression:
                    res['type'] = res_triple[2]
                elif res_triple[1] == GraphEntity.iri_has_sequence_identifier:
                    res['sequence_identifier'] = str(res_triple[2])
                elif res_triple[1] == GraphEntity.iri_part_of:
                    res['container'] = res_triple[2]
            ress.append(res)
        for res in ress:
            if res['res'] is not None:
                if res['type'] == GraphEntity.iri_journal_issue and res['container'] == venue_iri:
                    content['issue'].setdefault(res['sequence_identifier'], dict())
                    content['issue'][res['sequence_identifier']]['id'] = res['res']
                elif res['type'] == GraphEntity.iri_journal_volume:
                    content['volume'].setdefault(res['sequence_identifier'], dict())
                    content['volume'][res['sequence_identifier']]['id'] = res['res']
                    content['volume'][res['sequence_identifier']]['issue'] = self.__retrieve_issues_by_volume(URIRef(f"{self.base_iri}/br/{res['res']}"))
        return content

    def __retrieve_issues_by_volume(self, res:URIRef) -> dict:
        content = dict()
        for triple in self.local_g.triples((None, GraphEntity.iri_part_of, res)):
            for res_triple in self.local_g.triples((triple[0], None, None)):
                if res_triple[1] == GraphEntity.iri_has_sequence_identifier:
                    content.setdefault(str(res_triple[2]), dict())
                    content[str(res_triple[2])]['id'] = res_triple[0].replace(f'{self.base_iri}/br/', '')
        return content
    
    def retrieve_ra_sequence_from_br_meta(self, metaid: str, col_name: str) -> List[Dict[str, tuple]]:
        '''
        Given a bibliographic resource's MetaID and a field name, it returns its agent roles and responsible agents in the correct order according to the specified field.
        The output has the following format: ::
            [
                {METAID_AR_1: (NAME_RA_1, [(METAID_ID_RA_1, LITERAL_VALUE_ID_RA_1)], METAID_RA_1)}, 
                {METAID_AR_2: (NAME_RA_2, [(METAID_ID_RA_2, LITERAL_VALUE_ID_RA_2)], METAID_RA_2)}, 
                {METAID_AR_N: (NAME_RA_N, [(METAID_ID_RA_N, LITERAL_VALUE_ID_RA_N)], METAID_RA_N)}, 
            ]
            [
                {'5343': ('Hodge, James G.', [], '3316')}, 
                {'5344': ('Anderson, Evan D.', [], '3317')}, 
                {'5345': ('Kirsch, Thomas D.', [], '3318')}, 
                {'5346': ('Kelen, Gabor D.', [('4278', 'orcid:0000-0002-3236-8286')], '3319')}
            ]   
        :params metaid: a MetaID
        :type meta_id: str
        :params col_name: a MetaID
        :type col_name: str
        :returns: List[Dict[str, tuple]] -- the output is a list of three-elements tuples. Each tuple's first and third elements are the MetaIDs of an agent role and responsible agent related to the specified bibliographic resource. The second element is a two-elements tuple, where the first element is the MetaID of the identifier of the responsible agent. In contrast, the second one is the literal value of that id.
        '''
        if col_name == 'author':
            role = GraphEntity.iri_author
        elif col_name == 'editor':
            role = GraphEntity.iri_editor
        else:
            role = GraphEntity.iri_publisher
        
        metaid_uri = URIRef(f'{self.base_iri}/br/{str(metaid)}')
        dict_ar = dict()
        changes_made = False
        
        for triple in self.local_g.triples((metaid_uri, GraphEntity.iri_is_document_context_for, None)):
            for ar_triple in self.local_g.triples((triple[2], None, None)):
                if ar_triple[2] == role:
                    role_value = str(triple[2]).replace(f'{self.base_iri}/ar/', '')
                    next_role = ''
                    for relevant_ar_triple in self.local_g.triples((triple[2], None, None)):                            
                        if relevant_ar_triple[1] == GraphEntity.iri_has_next:
                            next_role = str(relevant_ar_triple[2]).replace(f'{self.base_iri}/ar/', '')
                        elif relevant_ar_triple[1] == GraphEntity.iri_is_held_by:
                            ra = str(relevant_ar_triple[2]).replace(f'{self.base_iri}/ra/', '')
                    dict_ar[role_value] = {'next': next_role, 'ra': ra}
        
        initial_dict_ar = dict_ar.copy()

        # Detect and handle duplicated RA
        ra_to_ars = {}
        for ar, details in dict_ar.items():
            ra = details['ra']
            if ra not in ra_to_ars:
                ra_to_ars[ra] = []
            ra_to_ars[ra].append(ar)

        # Identify and delete duplicate ARs
        ar_to_delete_list = []
        for ra, ars in ra_to_ars.items():
            if len(ars) > 1:
                # Keep the first AR and delete the rest
                for ar_to_delete in ars[1:]:
                    meta_editor = MetaEditor(meta_config=self.meta_config_path, resp_agent='https://w3id.org/oc/meta/prov/pa/1', save_queries=True)
                    meta_editor.delete(res=f"{self.base_iri}/ar/{ar_to_delete}")
                    ar_to_delete_list.append(ar_to_delete)
                    changes_made = True
        
        for ar in ar_to_delete_list:
            del dict_ar[ar]

        # Check for ARs that have themselves as 'next' and remove the 'next' relationship
        for ar, details in dict_ar.items():
            if details['next'] == ar:
                meta_editor = MetaEditor(meta_config=self.meta_config_path, resp_agent='https://w3id.org/oc/meta/prov/pa/1', save_queries=True)
                meta_editor.delete(res=f"{self.base_iri}/ar/{ar}", property=str(GraphEntity.iri_has_next))
                dict_ar[ar]['next'] = ''
                changes_made = True

        # Remove invalid 'next' references
        for role, details in list(dict_ar.items()):
            if details['next'] and details['next'] not in dict_ar:
                dict_ar[role]['next'] = ''
                changes_made = True

        # Find the start_role by excluding all roles that are "next" for others from the set of all roles.
        all_roles = set(dict_ar.keys())
        roles_with_next = set(details['next'] for details in dict_ar.values() if details['next'])
        start_role_candidates = all_roles - roles_with_next
        # Handle the edge cases for start role determination   

        MAX_ITERATIONS = 1000  # Numero massimo di iterazioni permesse
        SAFETY_TIMER = 3600  # Timer di sicurezza di 1 ora (in secondi)

        if len(all_roles) == 0:
            return []
        elif len(start_role_candidates) != 1:
            # If more than one start candidate exists or none exist in a multi-role situation, resolve automatically
            chains = []
            for start_candidate in start_role_candidates:
                current_role = start_candidate
                chain = []
                visited_roles = set()
                iteration_count = 0
                while current_role and current_role not in visited_roles and iteration_count < MAX_ITERATIONS:
                    visited_roles.add(current_role)
                    ra_info = self.retrieve_ra_from_meta(dict_ar[current_role]['ra'])[0:2]
                    ra_tuple = ra_info + (dict_ar[current_role]['ra'],)
                    chain.append({current_role: ra_tuple})
                    current_role = dict_ar[current_role]['next']
                    iteration_count += 1

                if iteration_count == MAX_ITERATIONS:
                    print(f"Possible infinite loop detected for BR: {metaid}")
                    print("Starting safety timer. Please stop the process if needed.")
                    sleep(SAFETY_TIMER)
                    return []  # Ritorna una lista vuota dopo il timer

                chains.append(chain)
            # Sort chains by length, then by the lowest sequential number of the starting role            
            chains.sort(key=lambda chain: (-len(chain), get_resource_number(f'{self.base_iri}/ar/{list(chain[0].keys())[0]}')))
            try:
                ordered_ar_list = chains[0]
            except Exception as e:
                print(f"\nProcessing BR: {metaid} for column: {col_name}")
                print(f"Initial dict_ar: {dict_ar}")
                print(f"All roles: {all_roles}")
                print(f"Start role candidates: {start_role_candidates}")
                print(f"Roles with next: {roles_with_next}")
                print(f"Error occurred while sorting or selecting chains: {str(e)}")
                print(f"Chains at time of error: {chains}")
                raise
            for chain in chains[1:]:
                for ar_dict in chain:
                    for ar in ar_dict.keys():
                        meta_editor = MetaEditor(meta_config=self.meta_config_path, resp_agent='https://w3id.org/oc/meta/prov/pa/1', save_queries=True)
                        meta_editor.delete(res=f"{self.base_iri}/ar/{ar}")
                        changes_made = True
        else:
            start_role = start_role_candidates.pop()
            # Follow the "next" chain from the start_role to construct an ordered list.
            ordered_ar_list = []
            current_role = start_role
            while current_role:
                ra_info = self.retrieve_ra_from_meta(dict_ar[current_role]['ra'])[0:2]
                ra_tuple = ra_info + (dict_ar[current_role]['ra'],)
                ordered_ar_list.append({current_role: ra_tuple})
                current_role = dict_ar[current_role]['next']

        final_chain = [list(ar_dict.keys())[0] for ar_dict in ordered_ar_list]

        # Fill gaps in the AR chain
        for i in range(len(final_chain) - 1):
            current_ar = final_chain[i]
            next_ar = final_chain[i + 1]
            if dict_ar[current_ar]['next'] != next_ar:
                meta_editor = MetaEditor(meta_config=self.meta_config_path, resp_agent='https://w3id.org/oc/meta/prov/pa/1', save_queries=True)
                meta_editor.update_property(
                    res=f"{self.base_iri}/ar/{current_ar}",
                    property=str(GraphEntity.iri_has_next),
                    new_value=URIRef(f"{self.base_iri}/ar/{next_ar}")
                )
                dict_ar[current_ar]['next'] = next_ar
                changes_made = True

        # Ensure the last AR doesn't have a 'next' relationship
        last_ar = final_chain[-1]
        if dict_ar[last_ar]['next']:
            meta_editor = MetaEditor(meta_config=self.meta_config_path, resp_agent='https://w3id.org/oc/meta/prov/pa/1', save_queries=True)
            meta_editor.delete(res=f"{self.base_iri}/ar/{last_ar}", property=GraphEntity.iri_has_next)
            dict_ar[last_ar]['next'] = ''
            changes_made = True

        if changes_made:
            print(f"\nChanges made to AR chain for BR: {metaid}")
            print(f"Initial AR chain: {initial_dict_ar}")
            print(f"Final AR chain: {dict_ar}")
            print(f"Final ordered AR list: {ordered_ar_list}\n")
        
        return ordered_ar_list

    def retrieve_re_from_br_meta(self, metaid:str) -> Tuple[str, str]:
        '''
            Given a bibliographic resource's MetaID, it returns its resource embodiment's MetaID and pages.
            The output has the following format: ::

                (METAID, PAGES)
                ('2011', '391-397')

            :params metaid: a bibliographic resource's MetaID
            :type meta_id: str
            :returns: Tuple[str, str] -- the output is a two-elements tuple, where the first element is the MetaID of the resource embodiment, and the second is a pages' interval. 
        '''
        metaid_uri = URIRef(f'{self.base_iri}/br/{str(metaid)}')
        re_uri = None
        starting_page = None
        ending_page = None
        for triple in self.local_g.triples((metaid_uri, GraphEntity.iri_embodiment, None)):
            re_uri = triple[2].replace(f'{self.base_iri}/re/', '')
            for re_triple in self.local_g.triples((triple[2], None, None)):
                if re_triple[1] == GraphEntity.iri_starting_page:
                    starting_page = str(re_triple[2])
                elif re_triple[1] == GraphEntity.iri_ending_page:
                    ending_page = str(re_triple[2])
        if re_uri:
            if starting_page and ending_page:
                pages = f'{starting_page}-{ending_page}'
            elif starting_page and not ending_page:
                pages = f'{starting_page}-{starting_page}'
            elif not starting_page and ending_page:
                pages = f'{ending_page}-{ending_page}'
            elif not starting_page and not ending_page:
                pages = ''
            return re_uri, pages

    def retrieve_br_info_from_meta(self, metaid: str) -> dict:
        '''
        Given a bibliographic resource's MetaID, it returns all the information about that resource.
        The output has the following format: ::

            {
                'pub_date': PUB_DATE, 
                'type': TYPE, 
                'page': (METAID, PAGES), 
                'issue': ISSUE, 
                'volume': VOLUME, 
                'venue': VENUE
            }
            {
                'pub_date': '2006-02-27', 
                'type': 'journal article', 
                'page': ('2011', '391-397'), 
                'issue': '4', 
                'volume': '166', 
                'venue': 'Archives Of Internal Medicine [omid:br/4387]'
            }

        :param metaid: a bibliographic resource's MetaID
        :type metaid: str
        :returns: dict -- the output is a dictionary including the publication date, type, page, issue, volume, and venue of the specified bibliographic resource.
        '''

        venue_iris = [
            GraphEntity.iri_archival_document,
            GraphEntity.iri_journal,
            GraphEntity.iri_book,
            GraphEntity.iri_book_series,
            GraphEntity.iri_series,
            GraphEntity.iri_academic_proceedings,
            GraphEntity.iri_proceedings_series,
            GraphEntity.iri_reference_book,
            GraphEntity.iri_series,

            GraphEntity.iri_expression
        ]

        def extract_identifiers(entity_uri):
            identifiers = [f"omid:{entity_uri.replace(f'{self.base_iri}/', '')}"]
            for id_triple in self.local_g.triples((entity_uri, GraphEntity.iri_has_identifier, None)):
                id_obj = id_triple[2]
                scheme = value = None
                for detail_triple in self.local_g.triples((id_obj, None, None)):
                    if detail_triple[1] == GraphEntity.iri_uses_identifier_scheme:
                        scheme = str(detail_triple[2])
                    elif detail_triple[1] == GraphEntity.iri_has_literal_value:
                        value = str(detail_triple[2])
                if scheme and value:
                    scheme = scheme.replace(GraphEntity.DATACITE, '')
                    identifiers.append(f"{scheme}:{value}")
            return identifiers

        metaid = str(metaid)
        metaid_uri = URIRef(f'{self.base_iri}/br/{metaid}') if self.base_iri not in metaid else URIRef(metaid)
        res_dict = {
            'pub_date': '',
            'type': '',
            'page': self.retrieve_re_from_br_meta(metaid),
            'issue': '',
            'volume': '',
            'venue': ''
        }

        for triple in self.local_g.triples((metaid_uri, None, None)):
            predicate, obj = triple[1], triple[2]

            if predicate == GraphEntity.iri_has_publication_date:
                res_dict['pub_date'] = str(obj)
            elif predicate == RDF.type and obj != GraphEntity.iri_expression:
                res_dict['type'] = self._type_it(obj)
            elif predicate == GraphEntity.iri_has_sequence_identifier:
                for inner_triple in self.local_g.triples((metaid_uri, None, None)):
                    inner_obj = inner_triple[2]
                    if inner_obj == GraphEntity.iri_journal_issue:
                        res_dict['issue'] = str(triple[2])
                    elif inner_obj == GraphEntity.iri_journal_volume:
                        res_dict['volume'] = str(triple[2])
            elif predicate == GraphEntity.iri_part_of:
                for vvi_triple in self.local_g.triples((obj, None, None)):
                    vvi_obj = vvi_triple[2]
                    if vvi_obj == GraphEntity.iri_journal_issue:
                        for inner_vvi_triple in self.local_g.triples((obj, None, None)):
                            if inner_vvi_triple[1] == GraphEntity.iri_has_sequence_identifier:
                                res_dict['issue'] = str(inner_vvi_triple[2])
                    elif vvi_obj == GraphEntity.iri_journal_volume:
                        for inner_vvi_triple in self.local_g.triples((obj, None, None)):
                            if inner_vvi_triple[1] == GraphEntity.iri_has_sequence_identifier:
                                res_dict['volume'] = str(inner_vvi_triple[2])
                    elif vvi_obj in venue_iris:
                        for inner_vvi_triple in self.local_g.triples((obj, None, None)):
                            if inner_vvi_triple[1] == GraphEntity.iri_title:
                                venue_title = str(inner_vvi_triple[2])
                                venue_ids = extract_identifiers(obj)
                                res_dict['venue'] = f"{venue_title} [{' '.join(venue_ids)}]"

                    if vvi_triple[1] == GraphEntity.iri_part_of:
                        for vi_triple in self.local_g.triples((vvi_obj, None, None)):
                            vi_obj = vi_triple[2]
                            if vi_obj == GraphEntity.iri_journal_volume:
                                for inner_vvi_triple in self.local_g.triples((vvi_obj, None, None)):
                                    if inner_vvi_triple[1] == GraphEntity.iri_has_sequence_identifier:
                                        res_dict['volume'] = str(inner_vvi_triple[2])
                            elif vi_obj in venue_iris:
                                for inner_vvi_triple in self.local_g.triples((vvi_obj, None, None)):
                                    if inner_vvi_triple[1] == GraphEntity.iri_title:
                                        venue_title = str(inner_vvi_triple[2])
                                        venue_ids = extract_identifiers(vvi_obj)
                                        res_dict['venue'] = f"{venue_title} [{' '.join(venue_ids)}]"

                            if vi_triple[1] == GraphEntity.iri_part_of:
                                for venue_triple in self.local_g.triples((vi_obj, None, None)):
                                    if venue_triple[1] == GraphEntity.iri_title:
                                        venue_title = str(venue_triple[2])
                                        venue_ids = extract_identifiers(vi_obj)
                                        res_dict['venue'] = f"{venue_title} [{' '.join(venue_ids)}]"
        return res_dict

    @staticmethod
    def _type_it(br_type: URIRef) -> str:
        output_type = ''
        if br_type == GraphEntity.iri_archival_document:
            output_type = 'archival document'
        if br_type == GraphEntity.iri_book:
            output_type = 'book'
        if br_type == GraphEntity.iri_book_chapter:
            output_type = 'book chapter'
        if br_type == GraphEntity.iri_part:
            output_type = 'book part'
        if br_type == GraphEntity.iri_expression_collection:
            output_type = 'book section'
        if br_type == GraphEntity.iri_book_series:
            output_type = 'book series'
        if br_type == GraphEntity.iri_book_set:
            output_type = 'book set'
        if br_type == GraphEntity.iri_data_file:
            output_type = 'data file'
        if br_type == GraphEntity.iri_thesis:
            output_type = 'dissertation'
        if br_type == GraphEntity.iri_journal:
            output_type = 'journal'
        if br_type == GraphEntity.iri_journal_article:
            output_type = 'journal article'
        if br_type == GraphEntity.iri_journal_issue:
            output_type = 'journal issue'
        if br_type == GraphEntity.iri_journal_volume:
            output_type = 'journal volume'
        if br_type == GraphEntity.iri_proceedings_paper:
            output_type = 'proceedings article'
        if br_type == GraphEntity.iri_academic_proceedings:
            output_type = 'proceedings'
        if br_type == GraphEntity.iri_reference_book:
            output_type = 'reference book'
        if br_type == GraphEntity.iri_reference_entry:
            output_type = 'reference entry'
        if br_type == GraphEntity.iri_series:
            output_type = 'series'
        if br_type == GraphEntity.iri_report_document:
            output_type = 'report'
        if br_type == GraphEntity.iri_specification_document:
            output_type = 'standard'
        return output_type
    
    def retrieve_publisher_from_br_metaid(self, metaid:str):
        metaid_uri = URIRef(f'{self.base_iri}/br/{metaid}')
        publishers = set()
        for triple in self.local_g.triples((metaid_uri, None, None)):
            if triple[1] == GraphEntity.iri_is_document_context_for:
                for document_triple in self.local_g.triples((triple[2], None, None)):
                    if document_triple[2] == GraphEntity.iri_publisher:
                        publishers.add(triple[2])
            elif triple[1] == GraphEntity.iri_part_of:
                for inner_triple in self.local_g.triples((triple[2], None, None)):
                    if inner_triple[1] == GraphEntity.iri_is_document_context_for:
                        for document_triple in self.local_g.triples((inner_triple[2], None, None)):
                            if document_triple[2] == GraphEntity.iri_publisher:
                                publishers.add(inner_triple[2])
                    elif inner_triple[1] == GraphEntity.iri_part_of:
                        for inner_inner_triple in self.local_g.triples((inner_triple[2], None, None)):
                            if inner_inner_triple[1] == GraphEntity.iri_is_document_context_for:
                                for document_triple in self.local_g.triples((inner_inner_triple[2], None, None)):
                                    if document_triple[2] == GraphEntity.iri_publisher:
                                        publishers.add(inner_inner_triple[2])
        publishers_output = []
        for publisher_uri in publishers:
            pub_identifiers = []
            pub_name = None
            for triple in self.local_g.triples((publisher_uri, None, None)):
                if triple[1] == GraphEntity.iri_is_held_by:
                    pub_metaid = triple[2].replace(f'{self.base_iri}/', 'omid:')
                    pub_identifiers.append(pub_metaid)
                    for ra_triple in self.local_g.triples((triple[2], None, None)):
                        pub_schema = None
                        pub_literal = None
                        if ra_triple[1] == GraphEntity.iri_name:
                            pub_name = ra_triple[2]
                        elif ra_triple[1] == GraphEntity.iri_has_identifier:
                            for id_triple in self.local_g.triples((ra_triple[2], None, None)):
                                if id_triple[1] == GraphEntity.iri_uses_identifier_scheme:
                                    pub_schema = id_triple[2].replace(f'{str(GraphEntity.DATACITE)}', '')
                                elif id_triple[1] == GraphEntity.iri_has_literal_value:
                                    pub_literal = id_triple[2]
                        if pub_schema is not None and pub_literal is not None:
                            pub_id = f'{pub_schema}:{pub_literal}'
                            pub_identifiers.append(pub_id)
            if pub_name is not None:
                pub_full = f'{pub_name} [{" ".join(pub_identifiers)}]'
            else:
                pub_full = f'[{" ".join(pub_identifiers)}]'
            publishers_output.append(pub_full)
        return '; '.join(publishers_output)
            
    def get_everything_about_res(self, metavals: set, identifiers: set, vvis: set, max_depth: int = 10) -> None:
        BATCH_SIZE = 10
        def batch_process(input_set, batch_size):
            """Generator to split input data into smaller batches if batch_size is not None."""
            if batch_size is None:
                yield input_set
            else:
                for i in range(0, len(input_set), batch_size):
                    yield input_set[i:i + batch_size]

        def process_batch(subjects, cur_depth):
            """Process each batch of subjects up to the specified depth."""
            if not subjects or (max_depth and cur_depth > max_depth):
                return

            next_subjects = set()
            for batch in batch_process(list(subjects), BATCH_SIZE):
                query_prefix = f'''
                    SELECT ?s ?p ?o
                    WHERE {{
                        VALUES ?s {{ {' '.join([f"<{s}>" for s in batch])} }}
                        ?s ?p ?o.
                    }}'''
                
                result = self.__query(query_prefix)
                if result:
                    for row in result['results']['bindings']:
                        s = URIRef(row['s']['value'])
                        p = URIRef(row['p']['value'])
                        o = row['o']['value']
                        o_type = row['o']['type']
                        o_datatype = URIRef(row['o']['datatype']) if 'datatype' in row['o'] else None
                        o = URIRef(o) if o_type == 'uri' else Literal(lexical_or_value=o, datatype=o_datatype)
                        self.local_g.add((s, p, o))
                        if isinstance(o, URIRef) and p not in {RDF.type, GraphEntity.iri_with_role, GraphEntity.iri_uses_identifier_scheme}:
                            next_subjects.add(str(o))

            # Dopo aver processato tutti i batch di questo livello, procedi con il prossimo livello di profondità
            process_batch(next_subjects, cur_depth + 1)

        def get_initial_subjects_from_metavals(metavals):
            """Convert metavals to a set of subjects."""
            return {f"{self.base_iri}/{mid.replace('omid:', '')}" for mid in metavals}

        def get_initial_subjects_from_identifiers(identifiers):
            """Convert identifiers to a set of subjects based on batch queries."""
            subjects = set()
            for batch in batch_process(list(identifiers), BATCH_SIZE):
                if not batch:
                    continue

                if self.blazegraph_full_text_search:
                    # Processing for text search enabled databases
                    for identifier in batch:
                        scheme, literal = identifier.split(":", 1)
                        escaped_identifier = literal.replace('\\', '\\\\').replace('"', '\\"')
                        query = f'''
                            PREFIX bds: <http://www.bigdata.com/rdf/search#>
                            SELECT ?s WHERE {{
                                ?literal bds:search "{escaped_identifier}" ;
                                        bds:matchAllTerms "true" ;
                                        ^<{GraphEntity.iri_has_literal_value}> ?id.
                                ?id <{GraphEntity.iri_uses_identifier_scheme}> <{GraphEntity.DATACITE + scheme}>;
                                    ^<{GraphEntity.iri_has_identifier}> ?s .
                            }}
                        '''
                        result = self.__query(query)
                        for row in result['results']['bindings']:
                            subjects.add(str(row['s']['value']))
                elif self.virtuoso_full_text_search:
                    union_blocks = []
                    for identifier in batch:
                        scheme, literal = identifier.split(':', maxsplit=1)[0], identifier.split(':', maxsplit=1)[1]
                        escaped_literal = literal.replace('\\', '\\\\').replace('"', '\\"')
                        union_blocks.append(f"""
                            {{  
                                {{
                                    ?id <{GraphEntity.iri_has_literal_value}> "{escaped_literal}" .
                                }}
                                UNION
                                {{
                                    ?id <{GraphEntity.iri_has_literal_value}> "{escaped_literal}"^^<{XSD.string}> .
                                }}
                                ?id <{GraphEntity.iri_uses_identifier_scheme}> <{GraphEntity.DATACITE + scheme}> .
                                ?s <{GraphEntity.iri_has_identifier}> ?id .                                    
                            }}
                        """)
                    union_query = " UNION ".join(union_blocks)
                    query = f'''
                        SELECT ?s WHERE {{
                            {union_query}
                        }}
                    '''
                    result = self.__query(query)
                    for row in result['results']['bindings']:
                        subjects.add(str(row['s']['value']))
                else:
                    identifiers_values = []
                    for identifier in batch:
                        scheme, literal = identifier.split(':', maxsplit=1)[0], identifier.split(':', maxsplit=1)[1]
                        escaped_literal = literal.replace('\\', '\\\\').replace('"', '\\"')
                        identifiers_values.append(f"(<{GraphEntity.DATACITE + scheme}> \"{escaped_literal}\")")
                    identifiers_values_str = " ".join(identifiers_values)
                    query = f'''
                        SELECT DISTINCT ?s WHERE {{
                            VALUES (?scheme ?literal) {{ {identifiers_values_str} }}
                            ?id <{GraphEntity.iri_uses_identifier_scheme}> ?scheme .
                            ?id <{GraphEntity.iri_has_literal_value}> ?literalValue  .
                            FILTER(str(?literalValue) = str(?literal))
                            ?s <{GraphEntity.iri_has_identifier}> ?id .
                        }}
                    '''
                    result = self.__query(query)
                    for row in result['results']['bindings']:
                        subjects.add(str(row['s']['value']))
            return subjects

        def get_initial_subjects_from_vvis(vvis):
            """Convert vvis to a set of subjects based on batch queries."""
            subjects = set()
            for batch in batch_process(list(vvis), BATCH_SIZE):
                if not batch:
                    continue

                for volume, issue, venue_metaid in batch:
                    vvi_type = GraphEntity.iri_journal_issue if issue else GraphEntity.iri_journal_volume
                    query = f'''
                        SELECT ?s WHERE {{
                            ?s a <{vvi_type}>;
                                <{GraphEntity.iri_part_of}>+ <{self.base_iri}/{venue_metaid.replace("omid:", "")}>;
                                <{GraphEntity.iri_has_sequence_identifier}> "{issue if issue else volume}".
                        }}
                    '''
                    result = self.__query(query)
                    for row in result['results']['bindings']:
                        subjects.add(str(row['s']['value']))
            return subjects

        initial_subjects = set()

        if metavals:
            initial_subjects.update(get_initial_subjects_from_metavals(metavals))
        
        if identifiers:
            initial_subjects.update(get_initial_subjects_from_identifiers(identifiers))

        if vvis:
            initial_subjects.update(get_initial_subjects_from_vvis(vvis))
    
        # Now start the depth-based processing
        process_batch(initial_subjects, 0)

    def get_subgraph(self, res: str, graphs_dict: dict) -> Graph|None:
        if res in graphs_dict:
            return graphs_dict[res]
        subgraph = Graph()
        for triple in self.local_g.triples((res, None, None)):
            subgraph.add(triple)
        if len(subgraph):
            graphs_dict[res] = subgraph
            return subgraph


#!python
# Copyright 2022, Arcangelo Massari <arcangelo.massari@unibo.it>, Arianna Moretti <arianna.moretti4@unibo.it>
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.


import os
from argparse import ArgumentParser

from tqdm import tqdm

from oc_meta.lib.csvmanager import CSVManager
from oc_meta.lib.file_manager import get_csv_data, write_csv


def get_diff(meta_csv_path: str, new: str, destination: str) -> None:
    new_ids_set = CSVManager.load_csv_column_as_set(new, 'id')
    meta_ids_set = extract_ids_from_meta(meta_csv_path)
    diff = new_ids_set.difference(meta_ids_set)
    datalist = [{'id': id} for id in diff]
    if datalist:
        write_csv(destination, datalist)

def extract_ids_from_meta(meta_csv_path: str) -> set:
    meta_ids_set = set()
    filepaths = os.listdir(meta_csv_path)
    pbar = tqdm(total=len(filepaths))
    for filepath in filepaths:
        data = get_csv_data(os.path.join(meta_csv_path, filepath))
        for row in data:
            meta_ids_set.update([identifier.replace('doi:', '') for identifier in row['id'].split() if not identifier.startswith('omid:')])
        pbar.update()
    pbar.close()
    return meta_ids_set


if __name__ == '__main__': # pragma: no cover
    arg_parser = ArgumentParser('get_diff.py', description='Get a list of identifiers in the second list but not in the first one')
    arg_parser.add_argument('-m', '--meta', dest='meta_csv_path', required=True, help='OpenCitations Meta CSV output path')
    arg_parser.add_argument('-w', '--wanted', dest='wanted', required=True, help='The path to a list of identifiers i CSV format')
    arg_parser.add_argument('-d', '--destination', dest='destination', required=True, help='The output CSV path')
    args = arg_parser.parse_args()
    get_diff(args.meta_csv_path, args.wanted, args.destination)


#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (c) 2023 Arcangelo Massari <arcangelo.massari@unibo.it>
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED 'AS IS' AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE

import argparse
import gzip
import logging
import multiprocessing
import os
import re
import time
import uuid
from functools import lru_cache
from typing import Match
from zipfile import ZIP_DEFLATED, ZipFile

import orjson
import rdflib
from rdflib import ConjunctiveGraph, URIRef
from tqdm import tqdm

# Variable used in several functions
entity_regex: str = r"^(.+)/([a-z][a-z])/(0[1-9]+0)?((?:[1-9][0-9]*)|(?:\d+-\d+))$"
prov_regex: str = r"^(.+)/([a-z][a-z])/(0[1-9]+0)?((?:[1-9][0-9]*)|(?:\d+-\d+))/prov/([a-z][a-z])/([1-9][0-9]*)$"

logging.basicConfig(level=logging.ERROR, format='%(asctime)s - %(levelname)s - %(message)s')

@lru_cache(maxsize=1024)
def _get_match_cached(regex: str, group: int, string: str) -> str:
    match: Match = re.match(regex, string)
    if match is not None:
        return match.group(group)
    else:
        return ""

def get_base_iri(res: URIRef) -> str:
    string_iri: str = str(res)
    if "/prov/" in string_iri:
        return _get_match_cached(prov_regex, 1, string_iri)
    else:
        return _get_match_cached(entity_regex, 1, string_iri)

def get_short_name(res: URIRef) -> str:
    string_iri: str = str(res)
    if "/prov/" in string_iri:
        return _get_match_cached(prov_regex, 5, string_iri)
    else:
        return _get_match_cached(entity_regex, 2, string_iri)

def get_prov_subject_short_name(prov_res: URIRef) -> str:
    string_iri: str = str(prov_res)
    if "/prov/" in string_iri:
        return _get_match_cached(prov_regex, 2, string_iri)
    else:
        return ""  # non-provenance entities do not have a prov_subject!

def get_prefix(res: URIRef) -> str:
    string_iri: str = str(res)
    if "/prov/" in string_iri:
        return ""  # provenance entities cannot have a supplier prefix
    else:
        return _get_match_cached(entity_regex, 3, string_iri)

def get_prov_subject_prefix(prov_res: URIRef) -> str:
    string_iri: str = str(prov_res)
    if "/prov/" in string_iri:
        return _get_match_cached(prov_regex, 3, string_iri)
    else:
        return ""  # non-provenance entities do not have a prov_subject!

def get_count(res: URIRef) -> str:
    string_iri: str = str(res)
    if "/prov/" in string_iri:
        return _get_match_cached(prov_regex, 6, string_iri)
    else:
        return _get_match_cached(entity_regex, 4, string_iri)

def get_prov_subject_count(prov_res: URIRef) -> str:
    string_iri: str = str(prov_res)
    if "/prov/" in string_iri:
        return _get_match_cached(prov_regex, 4, string_iri)
    else:
        return ""  # non-provenance entities do not have a prov_subject!

def get_resource_number(res: URIRef) -> int:
    string_iri: str = str(res)
    if "/prov/" in string_iri:
        match = _get_match_cached(prov_regex, 4, string_iri)
    else:
        match = _get_match_cached(entity_regex, 4, string_iri)
    if not match:
        logging.warning(f"Could not extract resource number from URI: {string_iri}")
        return -1  # or some other default value
    try:
        return int(match)
    except ValueError:
        logging.error(f"Invalid resource number in URI: {string_iri}, extracted: {match}")
        return -1  # or some other default value

def find_local_line_id(res: URIRef, n_file_item: int = 1) -> int:
    cur_number: int = get_resource_number(res)

    cur_file_split: int = 0
    while True:
        if cur_number > cur_file_split:
            cur_file_split += n_file_item
        else:
            cur_file_split -= n_file_item
            break

    return cur_number - cur_file_split

def find_paths(res: URIRef, base_dir: str, base_iri: str, default_dir: str, dir_split: int,
               n_file_item: int, is_json: bool = True):
    """
    This function is responsible for looking for the correct JSON file that contains the data related to the
    resource identified by the variable 'string_iri'. This search takes into account the organisation in
    directories and files, as well as the particular supplier prefix for bibliographic entities, if specified.
    In case no supplier prefix is specified, the 'default_dir' (usually set to "_") is used instead.
    """
    string_iri: str = str(res)

    cur_number: int = get_resource_number(res)

    if cur_number == -1:
        # Handle the error case
        logging.error(f"Could not process URI: {string_iri}")
        return None, None  # or some default paths

    # Find the correct file number where to save the resources
    cur_file_split: int = 0
    while True:
        if cur_number > cur_file_split:
            cur_file_split += n_file_item
        else:
            break

    # The data have been split in multiple directories and it is not something related
    # with the provenance data of the whole corpus (e.g. provenance agents)
    if dir_split and not string_iri.startswith(base_iri + "prov/"):
        # Find the correct directory number where to save the file
        cur_split: int = 0
        while True:
            if cur_number > cur_split:
                cur_split += dir_split
            else:
                break

        if "/prov/" in string_iri:  # provenance file of a bibliographic entity
            subj_short_name: str = get_prov_subject_short_name(res)
            short_name: str = get_short_name(res)
            sub_folder: str = get_prov_subject_prefix(res)
            file_extension: str = '.json' if is_json else '.nq'
            if sub_folder == "":
                sub_folder = default_dir
            if sub_folder == "":
                sub_folder = "_"  # enforce default value

            cur_dir_path: str = base_dir + subj_short_name + os.sep + sub_folder + \
                os.sep + str(cur_split) + os.sep + str(cur_file_split) + os.sep + "prov"
            cur_file_path: str = cur_dir_path + os.sep + short_name + file_extension
        else:  # regular bibliographic entity
            short_name: str = get_short_name(res)
            sub_folder: str = get_prefix(res)
            file_extension: str = '.json' if is_json else '.nt'
            if sub_folder == "":
                sub_folder = default_dir
            if sub_folder == "":
                sub_folder = "_"  # enforce default value

            cur_dir_path: str = base_dir + short_name + os.sep + sub_folder + os.sep + str(cur_split)
            cur_file_path: str = cur_dir_path + os.sep + str(cur_file_split) + file_extension
    # Enter here if no split is needed
    elif dir_split == 0:
        if "/prov/" in string_iri:
            subj_short_name: str = get_prov_subject_short_name(res)
            short_name: str = get_short_name(res)
            sub_folder: str = get_prov_subject_prefix(res)
            file_extension: str = '.json' if is_json else '.nq'
            if sub_folder == "":
                sub_folder = default_dir
            if sub_folder == "":
                sub_folder = "_"  # enforce default value

            cur_dir_path: str = base_dir + subj_short_name + os.sep + sub_folder + \
                os.sep + str(cur_file_split) + os.sep + "prov"
            cur_file_path: str = cur_dir_path + os.sep + short_name + file_extension
        else:
            short_name: str = get_short_name(res)
            sub_folder: str = get_prefix(res)
            file_extension: str = '.json' if is_json else '.nt'
            if sub_folder == "":
                sub_folder = default_dir
            if sub_folder == "":
                sub_folder = "_"  # enforce default value

            cur_dir_path: str = base_dir + short_name + os.sep + sub_folder
            cur_file_path: str = cur_dir_path + os.sep + str(cur_file_split) + file_extension
    # Enter here if the data is about a provenance agent, e.g. /corpus/prov/
    else:
        short_name: str = get_short_name(res)
        prefix: str = get_prefix(res)
        count: str = get_count(res)
        file_extension: str = '.json' if is_json else '.nq'

        cur_dir_path: str = base_dir + short_name
        cur_file_path: str = cur_dir_path + os.sep + prefix + count + file_extension

    return cur_dir_path, cur_file_path

def store(triples, graph_identifier, stored_g: ConjunctiveGraph) -> ConjunctiveGraph:
    for triple in triples:
        stored_g.add((triple[0], triple[1], triple[2], graph_identifier))
    return stored_g

def store_in_file(cur_g: ConjunctiveGraph, cur_file_path: str, zip_output: bool) -> None:
    dir_path = os.path.dirname(cur_file_path)
    if not os.path.exists(dir_path):
        os.makedirs(dir_path, exist_ok=True)

    cur_json_ld = orjson.loads(cur_g.serialize(format="json-ld"))

    if zip_output:
        with ZipFile(cur_file_path, mode="w", compression=ZIP_DEFLATED, allowZip64=True) as zip_file:
            json_str = orjson.dumps(cur_json_ld).decode('utf-8')
            zip_file.writestr(os.path.basename(cur_file_path.replace('.zip', '.json')), json_str)
    else:
        with open(cur_file_path, 'wb') as f:
            f.write(orjson.dumps(cur_json_ld))

def load_graph(file_path: str, cur_format: str = 'json-ld'):
    loaded_graph = ConjunctiveGraph()
    if file_path.endswith('.zip'):
        with ZipFile(file=file_path, mode="r", compression=ZIP_DEFLATED, allowZip64=True) as archive:
            for zf_name in archive.namelist():
                with archive.open(zf_name) as f:
                    if cur_format == "json-ld":
                        json_ld_file = orjson.loads(f.read())
                        if isinstance(json_ld_file, dict):
                            json_ld_file = [json_ld_file]
                        for json_ld_resource in json_ld_file:
                            loaded_graph.parse(data=orjson.dumps(json_ld_resource).decode('utf-8'), format=cur_format)
                    else:
                        loaded_graph.parse(file=f, format=cur_format)
    else:
        with open(file_path, 'rb', encoding='utf-8') as f:
            if cur_format == "json-ld":
                json_ld_file = orjson.loads(f.read())
                if isinstance(json_ld_file, dict):
                    json_ld_file = [json_ld_file]
                for json_ld_resource in json_ld_file:
                    loaded_graph.parse(data=orjson.dumps(json_ld_resource).decode('utf-8'), format=cur_format)
            else:
                loaded_graph.parse(file=f, format=cur_format)

    return loaded_graph

def process_graph(context, graph_identifier, output_root, base_iri, file_limit, item_limit, zip_output):
    modifications_by_file = {}
    triples = 0
    unique_id = generate_unique_id()

    for triple in context:
        triples += len(triple)
        entity_uri = triple[0]
        _, cur_file_path = find_paths(entity_uri, output_root, base_iri, '_', file_limit, item_limit, True)
        if cur_file_path is None:
            logging.warning(f"Skipping triple due to invalid URI: {entity_uri}")
            continue
        
        # Estrai il nome base del file (numero) e aggiungi l'ID unico
        base_name = os.path.splitext(os.path.basename(cur_file_path))[0]
        new_file_name = f"{base_name}_{unique_id}"
        
        cur_file_path = os.path.join(os.path.dirname(cur_file_path), new_file_name + ('.zip' if zip_output else '.json'))

        if cur_file_path not in modifications_by_file:
            modifications_by_file[cur_file_path] = {
                "graph_identifier": graph_identifier,
                "triples": []
            }
        modifications_by_file[cur_file_path]["triples"].append(triple)

    for file_path, data in modifications_by_file.items():
        stored_g = load_graph(file_path) if os.path.exists(file_path) else ConjunctiveGraph()
        stored_g = store(data["triples"], data["graph_identifier"], stored_g)
        store_in_file(stored_g, file_path, zip_output)
    return triples

def merge_files(output_root, base_file_name, file_extension, zip_output):
    """Funzione per fondere i file generati dai diversi processi"""
    files_to_merge = [f for f in os.listdir(output_root) if f.startswith(base_file_name) and f.endswith(file_extension)]
    
    merged_graph = ConjunctiveGraph()

    for file_path in files_to_merge:
        cur_full_path = os.path.join(output_root, file_path)
        loaded_graph = load_graph(cur_full_path)
        merged_graph += loaded_graph

    final_file_path = os.path.join(output_root, base_file_name + file_extension)
    store_in_file(merged_graph, final_file_path, zip_output)

def merge_files_in_directory(directory, zip_output, stop_file):
    """Function to merge files in a specific directory"""
    if check_stop_file(stop_file):
        print("Stop file detected. Stopping merge process.")
        return

    files = [f for f in os.listdir(directory) if f.endswith('.zip' if zip_output else '.json')]
    
    # Group files by their base name (number without the unique ID)
    file_groups = {}
    for file in files:
        match = re.match(r'^(\d+)(?:_[^.]+)?\.', file)
        if match:
            base_name = match.group(1)
            if base_name not in file_groups:
                file_groups[base_name] = []
            file_groups[base_name].append(file)
    
    for base_file_name, files_to_merge in file_groups.items():
        if check_stop_file(stop_file):
            print("Stop file detected. Stopping merge process.")
            return

        # Only proceed with merging if there's at least one file with an underscore
        if not any('_' in file for file in files_to_merge):
            continue

        merged_graph = ConjunctiveGraph()

        for file_path in files_to_merge:
            cur_full_path = os.path.join(directory, file_path)
            loaded_graph = load_graph(cur_full_path)
            for context in loaded_graph.contexts():
                graph_identifier = context.identifier
                for triple in context:
                    merged_graph.add(triple + (graph_identifier,))
        
        final_file_path = os.path.join(directory, f"{base_file_name}" + ('.zip' if zip_output else '.json'))
        store_in_file(merged_graph, final_file_path, zip_output)

        # Remove the original files after merging
        for file_path in files_to_merge:
            if file_path != os.path.basename(final_file_path):
                os.remove(os.path.join(directory, file_path))

def generate_unique_id():
    return f"{int(time.time())}-{uuid.uuid4()}"

def merge_files_wrapper(args):
    directory, zip_output, stop_file = args
    merge_files_in_directory(directory, zip_output, stop_file)

def merge_all_files_parallel(output_root, zip_output, stop_file):
    """Function to merge files in parallel"""
    if check_stop_file(stop_file):
        print("Stop file detected. Stopping merge process.")
        return

    directories_to_process = []
    for root, dirs, files in os.walk(output_root):
        if any(f.endswith('.zip' if zip_output else '.json') for f in files):
            directories_to_process.append(root)

    with multiprocessing.Pool() as pool:
        list(tqdm(pool.imap(merge_files_wrapper, 
                            [(dir, zip_output, stop_file) for dir in directories_to_process]), 
                  total=len(directories_to_process), 
                  desc="Merging files in directories"))

def process_file_content(file_path, output_root, base_iri, file_limit, item_limit, zip_output, rdf_format):
    with gzip.open(file_path, 'rb') as f:
        data = f.read().decode('utf-8')
        graph = ConjunctiveGraph()
        try:
            graph.parse(data=data, format=rdf_format)
        except rdflib.exceptions.ParserError as e:
            logging.error(f"Failed to parse {file_path}: {e}")
            return

    for context in graph.contexts():
        graph_identifier = context.identifier
        process_graph(context, graph_identifier, output_root, base_iri, file_limit, item_limit, zip_output)

def process_file_wrapper(args):
    file_path, output_root, base_iri, file_limit, item_limit, zip_output, rdf_format, cache_file, stop_file = args
    if check_stop_file(stop_file):
        return
    if not is_file_processed(file_path, cache_file):
        process_file_content(file_path, output_root, base_iri, file_limit, item_limit, zip_output, rdf_format)
        mark_file_as_processed(file_path, cache_file)

def process_chunk(chunk, output_root, base_iri, file_limit, item_limit, zip_output, rdf_format, cache_file, stop_file):
    with multiprocessing.Pool() as pool:
        list(tqdm(pool.imap(process_file_wrapper, 
                            [(file_path, output_root, base_iri, file_limit, item_limit, zip_output, rdf_format, cache_file, stop_file) 
                             for file_path in chunk]), 
                  total=len(chunk), 
                  desc="Processing files"))

def create_cache_file(cache_file):
    if cache_file:
        if not os.path.exists(cache_file):
            with open(cache_file, 'w', encoding='utf8') as f:
                pass  # Create an empty file
    else:
        logging.info("No cache file specified. Skipping cache creation.")

def is_file_processed(file_path, cache_file):
    if not cache_file:
        return False
    with open(cache_file, 'r', encoding='utf8') as f:
        processed_files = f.read().splitlines()
    return file_path in processed_files

def mark_file_as_processed(file_path, cache_file):
    if cache_file:
        with open(cache_file, 'a', encoding='utf8') as f:
            f.write(f"{file_path}\n")
    else:
        logging.debug(f"No cache file specified. Skipping marking {file_path} as processed.")

def check_stop_file(stop_file):
    return os.path.exists(stop_file)

def main():
    parser = argparse.ArgumentParser(description="Process gzipped input files into OC Meta RDF")
    parser.add_argument('input_folder', type=str, help='Input folder containing gzipped input files')
    parser.add_argument('output_root', type=str, help='Root folder for output OC Meta RDF files')
    parser.add_argument('--base_iri', type=str, default='https://w3id.org/oc/meta/', help='The base URI of entities on Meta')
    parser.add_argument('--file_limit', type=int, default=10000, help='Number of files per folder')
    parser.add_argument('--item_limit', type=int, default=1000, help='Number of items per file')
    parser.add_argument('-v', '--zip_output', default=True, dest='zip_output', action='store_true', help='Zip output json files')
    parser.add_argument('--input_format', type=str, default='jsonld', choices=['jsonld', 'nquads'], help='Format of the input files')
    parser.add_argument('--chunk_size', type=int, default=1000, help='Number of files to process before merging')
    parser.add_argument('--cache_file', type=str, default=None, help='File to store processed file names (optional)')
    parser.add_argument('--stop_file', type=str, default='./.stop', help='File to signal process termination')
    args = parser.parse_args()

    create_cache_file(args.cache_file)

    file_extension = '.nq.gz' if args.input_format == 'nquads' else '.jsonld.gz'
    rdf_format = 'nquads' if args.input_format == 'nquads' else 'json-ld'

    files_to_process = [os.path.join(args.input_folder, file) for file in os.listdir(args.input_folder) if file.endswith(file_extension)]
    chunks = [files_to_process[i:i + args.chunk_size] for i in range(0, len(files_to_process), args.chunk_size)]
    for i, chunk in enumerate(tqdm(chunks, desc="Processing chunks")):
        if check_stop_file(args.stop_file):
            print("Stop file detected. Gracefully terminating the process.")
            break
        print(f"\nProcessing chunk {i+1}/{len(chunks)}")
        process_chunk(chunk, args.output_root, args.base_iri, args.file_limit, args.item_limit, args.zip_output, rdf_format, args.cache_file, args.stop_file)
        print(f"Merging files for chunk {i+1}")
        merge_all_files_parallel(args.output_root, args.zip_output, args.stop_file)

    print("Processing complete")

if __name__ == "__main__":
    main()



#!python
# Copyright 2022, Arcangelo Massari <arcangelo.massari@unibo.it>, Arianna Moretti <arianna.moretti4@unibo.it>
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.


import os
from argparse import ArgumentParser

import yaml
from pebble import ProcessFuture, ProcessPool
from tqdm import tqdm

from oc_meta.plugins.fixer.ar_order import find_broken_roles


def task_done(task_output:ProcessFuture) -> None:
    PBAR.update()

if __name__ == '__main__': # pragma: no cover
    arg_parser = ArgumentParser('ar_order.py', description='Search for bibliographic resources where the order of roles is broken and correct this information')
    arg_parser.add_argument('-c', '--c', dest='config', required=True, help='OpenCitations Meta configuration file location')
    arg_parser.add_argument('-r', '--resp_agent', dest='resp_agent', required=True, help='Your ORCID URL')
    arg_parser.add_argument('-m', '--max_workers', dest='max_workers', required=False, default=1, type=int, help='Workers number')
    args = arg_parser.parse_args()
    with open(args.config, encoding='utf-8') as file:
        settings = yaml.full_load(file)
    rdf_dir = os.path.join(settings['output_rdf_dir'], 'rdf') + os.sep
    zip_output_rdf = settings['zip_output_rdf']
    br_files = [os.path.join(fold, file) for fold, _, files in os.walk(os.path.join(rdf_dir, 'br')) for file in files if file.endswith('.zip') and os.path.basename(fold) != 'prov']
    PBAR = tqdm(total=len(br_files))
    with ProcessPool(max_workers=args.max_workers, max_tasks=1) as executor:
        for filepath in br_files:
            future:ProcessFuture = executor.schedule(
                function=find_broken_roles, 
                args=(filepath, args.config, args.resp_agent, zip_output_rdf, True)) 
            future.add_done_callback(task_done)
    PBAR.close()


import argparse
import os
import zipfile
import lzma
from multiprocessing import Pool, cpu_count
from tqdm import tqdm

def process_file(zip_info_tuple):
    input_dir, output_dir, root, file = zip_info_tuple
    zip_path = os.path.join(root, file)
    with zipfile.ZipFile(zip_path, 'r') as zip_ref:
        for zip_info in zip_ref.infolist():
            with zip_ref.open(zip_info) as zip_file:
                file_content = zip_file.read()
                
                relative_path = os.path.relpath(root, input_dir)
                output_path = os.path.join(output_dir, relative_path)
                os.makedirs(output_path, exist_ok=True)
                
                lzma_file_path = os.path.join(output_path, zip_info.filename + '.xz')
                
                with lzma.open(lzma_file_path, 'w') as lzma_file:
                    lzma_file.write(file_content)

def zip_to_lzma(input_dir, output_dir):
    # Trova tutti i file zip nella directory di input
    zip_files = []
    for root, dirs, files in os.walk(input_dir):
        for file in files:
            if file.endswith('.zip'):
                zip_files.append((input_dir, output_dir, root, file))
    
    # Imposta il numero di processi pari al numero di CPU
    pool = Pool(processes=cpu_count())
    
    # Processa i file con una barra di progresso
    for _ in tqdm(pool.imap_unordered(process_file, zip_files), total=len(zip_files), desc="Processing files"):
        pass

    pool.close()
    pool.join()

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="Convert ZIP files to LZMA using multiprocessing")
    parser.add_argument("input_dir", help="Input directory containing ZIP files")
    parser.add_argument("output_dir", help="Output directory to save LZMA files")
    args = parser.parse_args()

    zip_to_lzma(args.input_dir, args.output_dir)



#!python
# Copyright (c) 2022 The OpenCitations Index Authors.
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.

import configparser
import json
import os
from os.path import join

import redis

from oc_meta.datasource.datasource import DataSource

# Read the Redis configuration file
config = configparser.ConfigParser(allow_no_value=True)
cur_path = os.path.dirname(os.path.abspath(__file__))
conf_file = join(cur_path, "config.ini")
config.read(conf_file)


class RedisDataSource(DataSource):
    def __init__(self, service):
        super().__init__(service)
        if service == "DB-META-RA":
            self._r =  redis.Redis(
                            host='127.0.0.1',
                            port=int(config.get('redis', 'port')),
                            db=(config.get('database 0', 'db')),
                            password=None,
                            decode_responses=True
                        )
        elif service == "DB-META-BR":
            self._r = redis.Redis(
                    host='127.0.0.1',
                    port=int(config.get('redis', 'port')),
                    db=(config.get('database 1', 'db')),
                    password=None,
                    decode_responses=True
                )
        else:
            raise ValueError

    def get(self, resource_id):
        redis_data = self._r.get(resource_id)
        if redis_data != None:
            if isinstance(redis_data, str):
                return redis_data
            else:
                return json.loads(redis_data)

    def mget(self, resources_id):
        return {
            resources_id[i]: json.loads(v) if not v is None else None
            for i, v in enumerate(self._r.mget(resources_id))
        }

    def set(self, resource_id, value):
        return self._r.set(resource_id, json.dumps(value))

    def mset(self, resources):
        return self._r.mset({k: json.dumps(v) for k, v in resources.items()})


#!python
# Copyright (c) 2022 The OpenCitations Index Authors.
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.

from abc import ABCMeta, abstractmethod


class DataSource(metaclass=ABCMeta):
    def __init__(self, service):
        self._service = service

    def new(self):
        return {"date": None, "valid": False, "issn": [], "orcid": []}

    @abstractmethod
    def get(self, resource_id):
        pass

    @abstractmethod
    def mget(self, resources_id):
        pass

    @abstractmethod
    def set(self, resource_id, value):
        pass

    @abstractmethod
    def mset(self, resources):
        pass


#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright 2023 Arcangelo Massari <arcangelo.massari@unibo.it>
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED 'AS IS' AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.


import os
from argparse import ArgumentParser

from tqdm import tqdm

from oc_meta.lib.file_manager import get_csv_data, write_csv

if __name__ == '__main__':
    arg_parser = ArgumentParser('remove_duplicated_ids.py', description='This script removes rows with duplicated ids from the Meta input CSVs')
    arg_parser.add_argument('-c', '--csv', dest='csv_dir', required=True,
                            help='The path to the directory containing the CSV files')
    arg_parser.add_argument('-o', '--output', dest='output_dir', required=True,
                            help='The path to the directory that will contain the CSVs in output')
    args = arg_parser.parse_args()
    data_found = set()
    ids_found = set()
    fieldnames = ['id', 'title', 'author', 'issue', 'volume', 'venue', 'pub_date', 'page', 'type', 'editor', 'publisher']
    pbar = tqdm(total=len(os.listdir(args.csv_dir)))
    for filename in os.listdir(args.csv_dir):
        data = get_csv_data(os.path.join(args.csv_dir, filename))
        output_data = []
        for row in data:
            row['id'] = ' '.join(sorted(row['id'].split()))
            row['author'] = '; '.join(sorted(row['author'].split('; ')))
            row['editor'] = '; '.join(sorted(row['editor'].split('; ')))
            data_in_row = ''.join([row[field] for field in fieldnames])
            if data_in_row not in data_found:
                output_data.append(row)
            data_found.add(data_in_row)
        write_csv(os.path.join(args.output_dir, filename), output_data, fieldnames)
        pbar.update()
    pbar.close()


#!python
# Copyright 2023, Arcangelo Massari <arcangelo.massari@unibo.it>, Arianna Moretti <arianna.moretti4@unibo.it>
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.

from argparse import ArgumentParser

import yaml
from rdflib import URIRef
from SPARQLWrapper import JSON, SPARQLWrapper
from tqdm import tqdm

from oc_meta.lib.file_manager import get_csv_data
from oc_meta.plugins.editor import MetaEditor

if __name__ == '__main__': # pragma: no cover
    arg_parser = ArgumentParser('duplicated_entities.py', description='Merge duplicated entities havng the same identifiers')
    arg_parser.add_argument('-i', '--csv', dest='csv_filepath', required=True, help='The path to a CSV file containing the list of entities to be merged. The CSV has two columns, "id" and "omid". The "id" column contains an identifier, while the "omid" column contains a list of omids separated by semicolon')
    arg_parser.add_argument('-e', '--entity_type', dest='entity_type', required=True, choices=['ra', 'br'], help='An entity type abbreviation')
    arg_parser.add_argument('-c', '--meta_config', dest='meta_config', required=True, help='OpenCitations Meta configuration file location')
    arg_parser.add_argument('-r', '--resp_agent', dest='resp_agent', required=True, help='Your ORCID URL')
    # arg_parser.add_argument('-ca', '--cache', dest='cache', required=True, help='Cache filepath')
    args = arg_parser.parse_args()
    with open(args.meta_config, encoding='utf-8') as file:
        settings = yaml.full_load(file)
    entity_type = args.entity_type
    csv_data = get_csv_data(args.csv_filepath)
    pbar = tqdm(total=len(csv_data))
    server = SPARQLWrapper(settings['triplestore_url'])
    meta_editor = MetaEditor(meta_config=args.meta_config, resp_agent=args.resp_agent)
    server.setReturnFormat(JSON)
    for entity in csv_data:
        identifier = entity['id']
        id_components = identifier.split(':', maxsplit=1)
        id_scheme = id_components[0]
        literal_value = id_components[1]
        omids = sorted([f'https://w3id.org/oc/meta/{omid}' for omid in entity['omid'].split('; ')])
        for omid in list(omids):
            query = f'''
                PREFIX datacite: <http://purl.org/spar/datacite/>
                PREFIX literal: <http://www.essepuntato.it/2010/06/literalreification/>
                ASK {{
                    <{omid}> datacite:hasIdentifier ?identifier.
                    ?identifier datacite:usesIdentifierScheme datacite:{id_scheme};
                                literal:hasLiteralValue "{literal_value}".
                }}
            '''
            server.setQuery(query)
            result = server.queryAndConvert()['boolean']
            if not result:
                print(f'The identifier {identifier} is not associated with the omid {omid}')
                omids.remove(omid)
        if len(omids) > 1:
            to_be_merged = [(omids[0], other) for other in omids[1:]]
            for merge_couple in to_be_merged:
                meta_editor.merge(URIRef(merge_couple[0]), URIRef(merge_couple[1]))
        pbar.update()
    pbar.close()


import os
import glob
import argparse
from tqdm import tqdm
from rdflib.plugins.sparql.parser import parseUpdate
from multiprocessing import Pool, cpu_count
from rdflib import Literal

# Function to process a file and extract quadruples to add and remove using rdflib
def process_sparql_file(filepath):
    quads_to_add = []
    quads_to_remove = []
    
    with open(filepath, 'r', encoding='utf-8') as file:
        query = file.read()
        try:
            parsed_query = parseUpdate(query)
            
            for update in parsed_query['request']:
                if update.name == 'InsertData':
                    for quadsNotTriples in update.quads.quadsNotTriples:
                        graph = quadsNotTriples.term
                        for quad in quadsNotTriples.triples:
                            subject, predicate, obj = quad
                            subject_str = f'<{subject}>'
                            predicate_str = f'<{predicate}>'
                            if 'string' in obj:
                                object_literal = Literal(lexical_or_value=obj.string, datatype=obj.datatype) if obj.datatype else Literal(lexical_or_value=obj.string)
                                object_str = object_literal.n3()
                            else:
                                object_str = f'<{obj}>'
                            quads_to_add.append((subject_str, predicate_str, object_str, f'<{graph}>'))
                elif update.name == 'DeleteData':
                    for quadsNotTriples in update.quads.quadsNotTriples:
                        graph = quadsNotTriples.term
                        for quad in quadsNotTriples.triples:
                            subject, predicate, obj = quad
                            subject_str = f'<{subject}>'
                            predicate_str = f'<{predicate}>'
                            if 'string' in obj:
                                object_literal = Literal(lexical_or_value=obj.string, datatype=obj.datatype) if obj.datatype else Literal(lexical_or_value=obj.string)
                                object_str = object_literal.n3()
                            else:
                                object_str = f'<{obj}>'
                            quads_to_remove.append((subject_str, predicate_str, object_str, f'<{graph}>'))
        except Exception as e:
            print(query, e)

    return quads_to_add, quads_to_remove

# Function to save quadruples in batches
def save_quads(quads, batch_size, prefix):
    os.makedirs(prefix, exist_ok=True)
    for i in tqdm(range(0, len(quads), batch_size), desc="Saving quads"):
        batch = quads[i:i+batch_size]
        with open(f'{prefix}/{i//batch_size + 1}.txt', 'w') as file:
            for quad in batch:
                file.write(' '.join(quad) + ' .\n')

# Function to save delete data queries in batches
def save_delete_data_queries(queries, batch_size, prefix):
    os.makedirs(prefix, exist_ok=True)
    for i in tqdm(range(0, len(queries), batch_size), desc="Saving delete queries"):
        batch = queries[i:i+batch_size]
        with open(f'{prefix}/{i//batch_size + 1}.txt', 'w') as file:
            file.write('DELETE DATA {\n')
            for graph in set(q[-1] for q in batch):
                file.write(f'  GRAPH {graph} {{\n')
                for query in batch:
                    if query[-1] == graph:
                        file.write('    ' + ' '.join(query[:-1]) + ' .\n')
                file.write('  }\n')
            file.write('}\n')

# Worker function for multiprocessing
def worker(filepath):
    return process_sparql_file(filepath)

# Main function to process all .sparql files in the directory using multiprocessing
def main(directory, output_directory):
    all_quads_to_add = []
    all_quads_to_remove = []
    
    filepaths = glob.glob(os.path.join(directory, '*.sparql'))
    
    # Create a pool of processes
    num_processes = cpu_count()
    pool = Pool(processes=num_processes)

    # Create a tqdm progress bar
    with tqdm(total=len(filepaths), desc="Processing files") as pbar:
        # Collect results from all processes
        for quads_add, quads_remove in pool.imap_unordered(worker, filepaths):
            all_quads_to_add.extend(quads_add)
            all_quads_to_remove.extend(quads_remove)
            pbar.update()

    # Close the pool
    pool.close()
    pool.join()

    # Create output subdirectories
    quads_to_add_dir = os.path.join(output_directory, 'quads_to_add')
    delete_data_queries_dir = os.path.join(output_directory, 'delete_data_queries')

    # Save quadruples to add in batches of 10,000
    save_quads(all_quads_to_add, 10000, quads_to_add_dir)
    
    # Save delete data queries in batches of 1,000
    save_delete_data_queries(all_quads_to_remove, 1000, delete_data_queries_dir)

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Process .sparql files and extract quadruples.')
    parser.add_argument('directory', type=str, help='The directory containing .sparql files')
    parser.add_argument('output_directory', type=str, help='The directory to save the output files')

    args = parser.parse_args()
    main(args.directory, args.output_directory)


import argparse
import os
import zipfile
from multiprocessing import Pool, cpu_count

from oc_meta.run.gen_info_dir import (get_prefix, get_resource_number,
                                      get_short_name)
from rdflib import ConjunctiveGraph, URIRef
from rdflib.namespace import PROV, RDF
from redis import Redis
from tqdm import tqdm


def process_zip_file(args):
    zip_file, redis_host, redis_port, redis_db = args
    redis_client = Redis(host=redis_host, port=redis_port, db=redis_db)
    missing_entities = []

    with zipfile.ZipFile(zip_file, 'r') as zip_ref:
        for file_name in zip_ref.namelist():
            with zip_ref.open(file_name) as entity_file:
                g = ConjunctiveGraph()
                g.parse(data=entity_file.read(), format='json-ld')
                
                for s, p, o in g.triples((None, RDF.type, PROV.Entity)):
                    prov_entity_uri = str(s)
                    entity_uri = prov_entity_uri.split('/prov/se/')[0]
                    supplier_prefix = get_prefix(entity_uri)
                    short_name = get_short_name(entity_uri)
                    resource_number = get_resource_number(entity_uri)
                    
                    expected_key = f"{short_name}:{supplier_prefix}:{resource_number}:se"
                    
                    if not redis_client.exists(expected_key):
                        print(f"\nEntità mancante trovata:")
                        print(f"URI: {entity_uri}")
                        print(f"Prov URI: {prov_entity_uri}")
                        print(f"Chiave Redis attesa: {expected_key}")
                        print("---")

                        missing_entities.append({
                            "URI": entity_uri,
                            "Prov URI": prov_entity_uri,
                            "Chiave Redis attesa": expected_key
                        })

    return missing_entities

def explore_provenance_files(root_path, redis_host, redis_port, redis_db):
    prov_zip_files = [os.path.join(dp, f) for dp, dn, filenames in os.walk(root_path) 
                      for f in filenames if f.endswith('.zip') and 'prov' in dp]

    args_list = [(zip_file, redis_host, redis_port, redis_db) for zip_file in prov_zip_files]
    
    num_processes = cpu_count()  # Usa tutti i core disponibili
    with Pool(processes=num_processes) as pool:
        results = list(tqdm(pool.imap(process_zip_file, args_list), total=len(args_list), desc="Processing provenance zip files"))

    all_missing_entities = [item for sublist in results for item in sublist]

    print(f"\nTotale entità mancanti trovate: {len(all_missing_entities)}")

def main():
    parser = argparse.ArgumentParser(description="Verifica la presenza di entità di provenance in Redis.")
    parser.add_argument("directory", type=str, help="Il percorso della directory da esplorare")
    parser.add_argument("--redis-host", type=str, default="localhost", help="L'host del server Redis")
    parser.add_argument("--redis-port", type=int, default=6379, help="La porta del server Redis")
    parser.add_argument("--redis-db", type=int, default=6, help="Il numero del database Redis da utilizzare")
    args = parser.parse_args()

    explore_provenance_files(args.directory, args.redis_host, args.redis_port, args.redis_db)

if __name__ == "__main__":
    main()


#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (c) 2022 Arcangelo Massari <arcangelo.massari@unibo.it>
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED 'AS IS' AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.


from __future__ import annotations

import os
from typing import List, Tuple

import yaml
from rdflib import URIRef
from tqdm import tqdm

from oc_meta.plugins.csv_generator.csv_generator import (find_file,
                                                         process_archive)
from oc_meta.plugins.editor import MetaEditor


def find_broken_roles(filepath: str, meta_config: str, resp_agent: str, zip_output_rdf: bool, merge_ra: bool = False):
    with open(meta_config, encoding='utf-8') as file:
        settings = yaml.full_load(file)
    rdf_dir = os.path.join(settings['output_rdf_dir'], 'rdf') + os.sep
    dir_split_number = settings['dir_split_number']
    items_per_file = settings['items_per_file']
    memory = dict()
    roles_in_br = process_archive(filepath, extract_roles_from_br)
    check_roles(roles_in_br, rdf_dir, dir_split_number, items_per_file, memory, meta_config, resp_agent, zip_output_rdf, merge_ra)

def extract_roles_from_br(br_data: list) -> list:
    all_ar = list()
    for graph in br_data:
        graph_data = graph['@graph']
        for br in graph_data:
            br_ars = list()
            if 'http://purl.org/spar/pro/isDocumentContextFor' in br:
                for ar_data in br['http://purl.org/spar/pro/isDocumentContextFor']:
                    br_ars.append(ar_data['@id'])
            if br_ars:
                all_ar.append(br_ars)
    return all_ar

def check_roles(roles_in_br: List[list], rdf_dir: str, dir_split_number: str, items_per_file: str, memory: dict, meta_config: str, resp_agent: str, zip_output_rdf: bool, merge_ra: bool = False) -> bool:
    order_changed_list = []
    meta_editor = MetaEditor(meta_config, resp_agent)
    for roles_list in roles_in_br:
        last_roles = {'author': {'has_next': dict(), 'ra': dict(), 'last': []}, 'editor': {'has_next': dict(), 'ra': dict(), 'last': []}, 'publisher': {'has_next': dict(), 'ra': dict(), 'last': []}}
        other_problems = {'author': {'self_next': False, 'multiple_has_next': False}, 'editor': {'self_next': False, 'multiple_has_next': False}, 'publisher': {'self_next': False, 'multiple_has_next': False}}
        to_be_merged = dict()
        for role in roles_list:
            ar_path = find_file(rdf_dir, dir_split_number, items_per_file, role, zip_output_rdf)
            role_data_tuple = process_archive(ar_path, get_ar_data, memory, role)
            if not role_data_tuple:
                meta_editor.sync_rdf_with_triplestore(role)
                role_data_tuple = process_archive(ar_path, get_ar_data, memory, role)
            if role_data_tuple:
                agent_role = role_data_tuple[0]
                has_next = role_data_tuple[1]
                resp_agent = role_data_tuple[2]
                agent_role = agent_role.split('http://purl.org/spar/pro/')[1]
                if resp_agent in last_roles[agent_role]['ra'] and merge_ra:
                    to_be_merged[last_roles[agent_role]['ra'][resp_agent]] = role
                    continue
                else:
                    last_roles[agent_role]['ra'][resp_agent] = role
                if len(has_next) == 0:
                    last_roles[agent_role]['last'].append(role)
                if any(next_role == role for next_role in has_next):
                    other_problems[agent_role]['self_next'] = True
                if len(has_next) > 1:
                    other_problems[agent_role]['multiple_has_next'] = True
                last_roles[agent_role]['has_next'][role] = has_next
        order_changed = fix_roles(last_roles, other_problems, to_be_merged, meta_config, resp_agent, merge_ra)
        order_changed_list.append(order_changed)
    return any(is_true for is_true in order_changed_list)

def fix_roles(last_roles: dict, other_problems: dict, to_be_merged: dict, meta_config: str, resp_agent: str, merge_ra: bool = False) -> bool:
    order_changed = False
    meta_editor = MetaEditor(meta_config, resp_agent)
    if merge_ra:
        for res, other in to_be_merged.items():
            meta_editor.merge(URIRef(res), URIRef(other))
    for role_type, role_data in last_roles.items():
        all_list = list(role_data['has_next'].keys())
        last_list = role_data['last']
        if (all_list and len(last_list) != 1) or other_problems[role_type]['self_next'] or other_problems[role_type]['multiple_has_next']:
            sorted_roles_list = sorted(all_list)
            for i, role in enumerate(sorted_roles_list):
                has_next_on_ts = last_roles[role_type]['has_next'][role][0] if last_roles[role_type]['has_next'][role] else ''
                if len(last_roles[role_type]['has_next'][role]) > 1:
                    meta_editor.delete(URIRef(role), 'has_next')
                    order_changed = True
                    last_roles[role_type]['has_next'][role] = []
                if i < len(sorted_roles_list) - 1:
                    if has_next_on_ts != sorted_roles_list[i+1]:
                        meta_editor.delete(URIRef(role), 'has_next')
                        order_changed = True
                elif i == len(sorted_roles_list) - 1:
                    if has_next_on_ts:
                        meta_editor.delete(URIRef(role), 'has_next')
                        order_changed = True
            for i, role in enumerate(sorted_roles_list):
                has_next_on_ts = last_roles[role_type]['has_next'][role][0] if last_roles[role_type]['has_next'][role] else ''
                if i < len(sorted_roles_list) - 1:
                    if has_next_on_ts != sorted_roles_list[i+1]:
                        meta_editor.update_property(URIRef(role), 'has_next', URIRef(sorted_roles_list[i+1]))
                        order_changed = True
    return order_changed

def get_ar_data(ar_data: list, ar_uri: str) -> Tuple[str, str, str]:
    for graph in ar_data:
        graph_data = graph['@graph']
        for agent in graph_data:
            if agent['@id'] == ar_uri:
                role = agent['http://purl.org/spar/pro/withRole'][0]['@id']
                ra = agent['http://purl.org/spar/pro/isHeldBy'][0]['@id']
                has_nexts = []
                if 'https://w3id.org/oc/ontology/hasNext' in agent:
                    for has_next in agent['https://w3id.org/oc/ontology/hasNext']:
                        has_nexts.append(has_next['@id'])
                    return role, has_nexts, ra
                return role, [], ra


import argparse
import multiprocessing
from multiprocessing import Process, Queue

import os
import time
from rdflib import ConjunctiveGraph, Literal, URIRef
from SPARQLWrapper import JSON, SPARQLWrapper
from tqdm import tqdm
import queue
from multiprocessing import Manager
from pebble import ProcessPool

def count_total_triples(endpoint):
    sparql = SPARQLWrapper(endpoint)
    sparql.setQuery("""
        SELECT (COUNT(*) AS ?totalTriples) WHERE {
            ?s ?p ?o .
        }
    """)
    sparql.setReturnFormat(JSON)
    try:
        results = sparql.query().convert()
        total_triples = int(results["results"]["bindings"][0]["totalTriples"]["value"])
        return total_triples
    except Exception as e:
        print(f"Errore nella query: {e}")
        return None

def query_triplestore_by_class(endpoint, class_uri, uri_prefix, supplier_prefix, offset, limit):
    sparql = SPARQLWrapper(endpoint)
    full_uri_prefix = f"{uri_prefix}/{supplier_prefix}"
    sparql.setQuery(f"""
        SELECT ?g ?s ?p ?o WHERE {{
            GRAPH ?g {{
                ?s a <{class_uri}>.
                FILTER (STRSTARTS(STR(?s), "{full_uri_prefix}")).
                ?s ?p ?o.
            }}
        }} LIMIT {limit} OFFSET {offset}
    """)
    sparql.setReturnFormat(JSON)
    results = sparql.query().convert()
    return results, bool(results["results"]["bindings"])

def convert_to_graph(results):
    g = ConjunctiveGraph()
    triple_count = 0
    for result in results["results"]["bindings"]:
        graph_uri = URIRef(result["g"]["value"])
        s = URIRef(result["s"]["value"])
        p = URIRef(result["p"]["value"])
        o_value = result["o"]["value"]
        if result["o"]["type"] == "uri":
            o = URIRef(o_value)
        else:
            datatype = result["o"].get("datatype")
            if datatype:
                o = Literal(o_value, datatype=URIRef(datatype))
            else:
                o = Literal(o_value)
        g.add((s, p, o, graph_uri))
        triple_count += 1
    return g, triple_count

def generate_supplier_prefixes():
    prefixes = []
    for i in range(1, 10000):
        if i % 10 != 0:
            prefix = f"06{i}0"
            prefixes.append(prefix)
    return prefixes

def process_task(class_uri, uri_prefix, supplier_prefixes, endpoint, output_folder, page_size, progress_queue):
    for supplier_prefix in supplier_prefixes:
        output_folder_prefix = os.path.join(output_folder, class_uri.split('/')[-1], supplier_prefix)
        os.makedirs(output_folder_prefix, exist_ok=True)
        file_count = 0
        offset = 0
        while True:
            results, has_data = query_triplestore_by_class(endpoint, class_uri, uri_prefix, supplier_prefix, offset, page_size)
            if has_data:
                graph, triple_count = convert_to_graph(results)
                output_filename = os.path.join(output_folder_prefix, f"{class_uri.split('/')[-1]}_{supplier_prefix}_output_{file_count}.jsonld")
                with open(output_filename, 'w', encoding='utf-8') as f:
                    jsonld_data = graph.serialize(format='json-ld', indent=None, ensure_ascii=False)
                    f.write(jsonld_data)
                progress_queue.put(triple_count)
            else:
                break
            offset += page_size
            file_count += 1
    progress_queue.put(None)

def progress_monitor(total_triples, progress_queue, class_uris):
    pbar = tqdm(total=total_triples)
    completed_processes = 0
    while completed_processes < len(class_uris):
        update = progress_queue.get()
        if update is None:
            completed_processes += 1
        else:
            pbar.update(update)
    pbar.close()

def main(endpoint, output_folder, page_size):
    class_uris = [
        "http://purl.org/spar/fabio/Expression",
        "http://purl.org/spar/fabio/Manifestation",
        "http://xmlns.com/foaf/0.1/Agent",
        "http://purl.org/spar/datacite/Identifier",
        "http://purl.org/spar/pro/RoleInTime"
    ]

    uri_prefixes = {
        "http://purl.org/spar/fabio/Expression": "https://w3id.org/oc/meta/br",
        "http://purl.org/spar/fabio/Manifestation": "https://w3id.org/oc/meta/re",
        "http://xmlns.com/foaf/0.1/Agent": "https://w3id.org/oc/meta/ra",
        "http://purl.org/spar/datacite/Identifier": "https://w3id.org/oc/meta/id",
        "http://purl.org/spar/pro/RoleInTime": "https://w3id.org/oc/meta/ar"
    }

    supplier_prefixes = generate_supplier_prefixes()

    total_triples = count_total_triples(endpoint)
    results = []

    total_triples = count_total_triples(endpoint)
    if total_triples is None:
        return

    with Manager() as manager:
        progress_queue = manager.Queue()
        monitor = Process(target=progress_monitor, args=(total_triples, progress_queue, class_uris))
        monitor.start()

        with ProcessPool() as pool:
            for class_uri in class_uris:
                uri_prefix = uri_prefixes[class_uri]
                pool.schedule(process_task, args=(class_uri, uri_prefix, supplier_prefixes, endpoint, output_folder, page_size, progress_queue))

        monitor.join()


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Download triples from a Blazegraph triplestore and save as JSON-LD.")
    parser.add_argument("--endpoint", required=True, help="SPARQL endpoint URL")
    parser.add_argument("--output", required=True, help="Folder path to save output JSON-LD files")
    parser.add_argument("--pagesize", type=int, default=10000, help="Number of triples per page (default: 1000)")

    args = parser.parse_args()
    main(args.endpoint, args.output, args.pagesize)


import argparse
import zipfile
from pathlib import Path
from rdflib import ConjunctiveGraph, URIRef
from tqdm import tqdm
import json
import multiprocessing

def process_zip_file(zip_path):
    """ Process a single zip file, read JSON-LD contents and check for multiple 'embodiment' properties. """
    graph = ConjunctiveGraph()
    targeted_subjects = {}

    with zipfile.ZipFile(zip_path, 'r') as zipf:
        for file in zipf.namelist():
            if file.endswith('.json'):
                with zipf.open(file) as f:
                    data = f.read()
                    graph.parse(data=data, format='json-ld')

    embodiment_uri = URIRef("http://purl.org/vocab/frbr/core#embodiment")
    for subject in graph.subjects(unique=True):
        embodiments = list(graph.objects(subject, embodiment_uri, unique=True))
        if len(embodiments) > 1:
            targeted_subjects[str(subject)] = [str(emb) for emb in embodiments]

    return targeted_subjects

def main(directory):
    """ Main function to process all zip files in the given directory excluding 'se.zip'. """
    path = Path(directory)
    all_zip_files = [f for f in path.rglob('*.zip') if f.name != 'se.zip']

    with multiprocessing.Pool() as pool:
        results = list(tqdm(pool.imap(process_zip_file, all_zip_files), total=len(all_zip_files)))

    all_targeted_subjects = {}

    for subjects in results:
        all_targeted_subjects.update(subjects)

    # Save the subjects with multiple embodiments to a JSON file
    with open('targeted_subjects.json', 'w') as f:
        json.dump(all_targeted_subjects, f)

    print(f"Subjects with multiple embodiments saved to 'targeted_subjects.json'.")

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Process .zip files to extract subjects with multiple "embodiment" properties.')
    parser.add_argument('--directory', type=str, required=True, help='Directory to scan for .zip files')
    args = parser.parse_args()
    
    main(args.directory)


import argparse
import concurrent.futures
import csv
import os
import traceback
from typing import Dict, List, Set

from oc_meta.plugins.editor import MetaEditor
from oc_ocdm.graph import GraphSet
from rdflib import URIRef
from SPARQLWrapper import SPARQLWrapper
from tqdm import tqdm


class EntityMerger:
    def __init__(self, meta_config: str, resp_agent: str, entity_types: List[str], stop_file_path: str, workers: int):
        self.meta_config = meta_config
        self.resp_agent = resp_agent
        self.entity_types = entity_types
        self.stop_file_path = stop_file_path
        self.workers = workers
        self.batch_size = 10

    @staticmethod
    def get_entity_type(entity_url: str) -> str:
        parts = entity_url.split('/')
        if 'oc' in parts and 'meta' in parts:
            try:
                return parts[parts.index('meta') + 1]
            except IndexError:
                return None
        return None

    @staticmethod
    def read_csv(csv_file: str) -> List[Dict]:
        data = []
        with open(csv_file, mode='r', newline='', encoding='utf-8') as file:
            csv_reader = csv.DictReader(file)
            for row in csv_reader:
                if 'Done' not in row:
                    row['Done'] = 'False'
                data.append(row)
        return data

    @staticmethod
    def write_csv(csv_file: str, data: List[Dict]):
        fieldnames = data[0].keys()
        with open(csv_file, mode='w', newline='', encoding='utf-8') as file:
            writer = csv.DictWriter(file, fieldnames=fieldnames)
            writer.writeheader()
            for row in data:
                writer.writerow(row)

    @staticmethod
    def count_csv_rows(csv_file: str) -> int:
        with open(csv_file, 'r', encoding='utf-8') as f:
            return sum(1 for _ in f) - 1

    def fetch_related_entities_batch(self, meta_editor: MetaEditor, merged_entities: List[str], 
                                surviving_entities: List[str], batch_size: int = 10) -> Set[URIRef]:
        """
        Fetch all related entities in batches and populate the relationship cache.
        
        Args:
            meta_editor: MetaEditor instance
            merged_entities: List of entities to be merged
            surviving_entities: List of surviving entities
            batch_size: Maximum number of entities to process in a single SPARQL query
            
        Returns:
            Set of all related entities
        """
        all_related_entities = set()
        
        for i in range(0, len(merged_entities), batch_size):
            batch_merged = merged_entities[i:i + batch_size]
            merged_clauses = []
            for entity in batch_merged:
                merged_clauses.extend([
                    f"{{?entity ?p <{entity}>}}", 
                    f"{{<{entity}> ?p ?entity}}"  
                ])
                
            if not merged_clauses:
                continue
                
            query = f"""
                PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
                PREFIX datacite: <http://purl.org/spar/datacite/>
                PREFIX pro: <http://purl.org/spar/pro/>
                SELECT DISTINCT ?entity WHERE {{
                    {{
                        {' UNION '.join(merged_clauses)}
                    }}
                    FILTER (?p != rdf:type) 
                    FILTER (?p != datacite:usesIdentifierScheme) 
                    FILTER (?p != pro:withRole)
                }}
            """
            
            sparql = SPARQLWrapper(meta_editor.endpoint)
            try:
                results = meta_editor.make_sparql_query_with_retry(sparql, query)
                for result in results["results"]["bindings"]:
                    if result['entity']['type'] == 'uri':
                        related_uri = URIRef(result['entity']['value'])
                        all_related_entities.add(related_uri)
                        
                        for entity in batch_merged:
                            entity_uri = URIRef(entity)
                            if entity_uri not in meta_editor.relationship_cache:
                                meta_editor.relationship_cache[entity_uri] = set()
                            meta_editor.relationship_cache[entity_uri].add(related_uri)
                        
            except Exception as e:
                print(f"Error fetching related entities for merged batch {i}-{i+batch_size}: {e}")
        
        for i in range(0, len(surviving_entities), batch_size):
            batch_surviving = surviving_entities[i:i + batch_size]
            surviving_clauses = []
            for entity in batch_surviving:
                surviving_clauses.append(
                    f"{{<{entity}> ?p ?entity}}"
                )
                
            if not surviving_clauses:
                continue
            
            query = f"""
                PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
                PREFIX datacite: <http://purl.org/spar/datacite/>
                PREFIX pro: <http://purl.org/spar/pro/>
                SELECT DISTINCT ?entity WHERE {{
                    {{
                        {' UNION '.join(surviving_clauses)}
                    }}
                    FILTER (?p != rdf:type) 
                    FILTER (?p != datacite:usesIdentifierScheme) 
                    FILTER (?p != pro:withRole)
                }}
            """
            
            sparql = SPARQLWrapper(meta_editor.endpoint)
            try:
                results = meta_editor.make_sparql_query_with_retry(sparql, query)
                for result in results["results"]["bindings"]:
                    if result['entity']['type'] == 'uri':
                        related_uri = URIRef(result['entity']['value'])
                        all_related_entities.add(related_uri)
                        
                        for entity in batch_surviving:
                            entity_uri = URIRef(entity)
                            if entity_uri not in meta_editor.relationship_cache:
                                meta_editor.relationship_cache[entity_uri] = set()
                            meta_editor.relationship_cache[entity_uri].add(related_uri)
                        
            except Exception as e:
                print(f"Error fetching related entities for surviving batch {i}-{i+batch_size}: {e}")
                
        return all_related_entities
    
    def should_stop_processing(self) -> bool:
        return os.path.exists(self.stop_file_path)
    
    def process_file(self, csv_file: str) -> str:
        """Process a single CSV file with cross-row batch processing"""
        data = self.read_csv(csv_file)
        meta_editor = MetaEditor(self.meta_config, self.resp_agent, save_queries=True)
        modified = False
        
        if self.should_stop_processing():
            return csv_file

        g_set = GraphSet(meta_editor.base_iri, custom_counter_handler=meta_editor.counter_handler)

        batch_merged_entities = []
        batch_surviving_entities = []
        rows_to_process = []

        for row in data:
            if row.get('Done') == 'True':
                continue

            entity_type = self.get_entity_type(row['surviving_entity'])
            if entity_type in self.entity_types:
                surviving_entity = row['surviving_entity']
                merged_entities = row['merged_entities'].split('; ')
                
                batch_surviving_entities.append(surviving_entity)
                batch_merged_entities.extend(merged_entities)
                rows_to_process.append((surviving_entity, merged_entities))

        if not rows_to_process:
            return csv_file

        all_related_entities = self.fetch_related_entities_batch(
            meta_editor,
            batch_merged_entities,
            batch_surviving_entities,
            self.batch_size
        )

        entities_to_import = all_related_entities.copy()
        entities_to_import.update(URIRef(e) for e in batch_surviving_entities)
        entities_to_import.update(URIRef(e) for e in batch_merged_entities)

        entities_to_import = {e for e in entities_to_import 
                            if not meta_editor.entity_cache.is_cached(e)}

        if entities_to_import:
            try:
                meta_editor.reader.import_entities_from_triplestore(
                    g_set=g_set,
                    ts_url=meta_editor.endpoint,
                    entities=list(entities_to_import),
                    resp_agent=meta_editor.resp_agent,
                    enable_validation=False,
                    batch_size=self.batch_size
                )
                
                for entity in entities_to_import:
                    meta_editor.entity_cache.add(entity)
                    
            except ValueError as e:
                print(f"Error importing entities: {e}")
                modified = True

        for surviving_entity, merged_entities in rows_to_process:
            surviving_uri = URIRef(surviving_entity)
            for merged_entity in merged_entities:
                try:
                    meta_editor.merge(g_set, surviving_uri, URIRef(merged_entity))
                    modified = True
                except ValueError:
                    continue

        if modified:
            for row in data:
                if row.get('Done') != 'True' and self.get_entity_type(row['surviving_entity']) in self.entity_types:
                    row['Done'] = 'True'
            
            meta_editor.save(g_set)
            self.write_csv(csv_file, data)

        return csv_file

    def process_folder(self, csv_folder: str):
        if os.path.exists(self.stop_file_path):
            os.remove(self.stop_file_path)

        csv_files = [os.path.join(csv_folder, file) 
                    for file in os.listdir(csv_folder) 
                    if file.endswith('.csv')]

        if self.workers > 4:
            csv_files = [file for file in csv_files 
                        if self.count_csv_rows(file) <= 10000]

        with concurrent.futures.ProcessPoolExecutor(max_workers=self.workers) as executor:
            futures = {}
            for csv_file in csv_files:
                if self.should_stop_processing():
                    break
                futures[executor.submit(self.process_file, csv_file)] = csv_file

            for future in tqdm(concurrent.futures.as_completed(futures), 
                             total=len(futures), 
                             desc="Overall Progress"):
                csv_file = futures[future]
                try:
                    future.result()
                except Exception as e:
                    error_trace = traceback.format_exc()
                    print(f"""
                        Error processing file {csv_file}:
                        Type: {type(e).__name__}
                        Details: {str(e)}
                        Full Traceback:
                        {error_trace}
                        Suggestion: This is an unexpected error. Please check the traceback for more details.
                    """)


def main():
    parser = argparse.ArgumentParser(description="Merge entities from CSV files in a folder.")
    parser.add_argument('csv_folder', type=str, 
                       help="Path to the folder containing CSV files")
    parser.add_argument('meta_config', type=str, 
                       help="Meta configuration string")
    parser.add_argument('resp_agent', type=str, 
                       help="Responsible agent string")
    parser.add_argument('--entity_types', nargs='+', 
                       default=['ra', 'br', 'id'], 
                       help="Types of entities to merge (ra, br, id)")
    parser.add_argument('--stop_file', type=str, 
                       default="stop.out", 
                       help="Path to the stop file")
    parser.add_argument('--workers', type=int, 
                       default=4, 
                       help="Number of parallel workers")
    
    args = parser.parse_args()

    merger = EntityMerger(
        meta_config=args.meta_config,
        resp_agent=args.resp_agent,
        entity_types=args.entity_types,
        stop_file_path=args.stop_file,
        workers=args.workers
    )
    
    merger.process_folder(args.csv_folder)


if __name__ == "__main__":
    main()


#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (c) 2022, Arcangelo Massari <arcangelo.massari@unibo.it>
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED 'AS IS' AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.


import os
from argparse import ArgumentParser

import yaml
from pebble import ProcessFuture, ProcessPool
from tqdm import tqdm

from oc_meta.lib.file_manager import pathoo
from oc_meta.plugins.csv_generator.csv_generator import (generate_csv,
                                                         process_archives,
                                                         process_br)


def task_done(task_output:ProcessFuture) -> None:
    PBAR.update()

if __name__ == '__main__': # pragma: no cover
    arg_parser = ArgumentParser('csv_generator.py', description='This script generates output CSVs from the OpenCitations Meta RDF dump')
    arg_parser.add_argument('-c', '--c', dest='config', required=True, help='OpenCitations Meta configuration file location')
    arg_parser.add_argument('-o', '--output_dir', dest='output_dir', required=True,
                            help='The output directory where the CSV files will be stores')
    arg_parser.add_argument('-t', '--threshold', dest='threshold', required=True, default=5000, type=int,
                            help='How many lines the CSV output must contain')
    arg_parser.add_argument('-m', '--max_workers', dest='max_workers', required=False, default=1, type=int, help='Workers number')
    arg_parser.add_argument('-w', '--wanted_schemas', dest='wanted_schemas', required=False, default=None, nargs='+', type=str, help='A list of wanted identifier schemes')
    args = arg_parser.parse_args()
    with open(args.config, encoding='utf-8') as file:
        settings = yaml.full_load(file)
    output_dir = args.output_dir
    wanted_schemas = args.wanted_schemas
    rdf_dir = os.path.join(settings['output_rdf_dir'], 'rdf') + os.sep
    dir_split_number = settings['dir_split_number']
    items_per_file = settings['items_per_file']
    resp_agent = settings['resp_agent']
    zip_output_rdf = settings['zip_output_rdf']
    if not os.path.exists(output_dir):
        pathoo(output_dir)
        process_archives(rdf_dir, 'br', output_dir, process_br, dir_split_number, items_per_file, zip_output_rdf, args.config, resp_agent, wanted_schemas, args.threshold)
    print('[csv_generator: INFO] Solving the OpenCitations Meta Identifiers recursively')
    filenames = os.listdir(output_dir)
    PBAR = tqdm(total=len(filenames))
    # for filename in filenames:
    #     generate_csv(filename, args.config, rdf_dir, dir_split_number, items_per_file, resp_agent, args.output_dir, zip_output_rdf)
    #     PBAR.update()
    with ProcessPool(max_workers=args.max_workers, max_tasks=1) as executor:
        for filename in filenames:
            future:ProcessFuture = executor.schedule(
                function=generate_csv,
                args=(filename, args.config, rdf_dir, dir_split_number, items_per_file, resp_agent, args.output_dir, zip_output_rdf))
            future.add_done_callback(task_done)
        PBAR.update()
    PBAR.close()


import os
from argparse import ArgumentParser

from tqdm import tqdm

from oc_meta.lib.file_manager import get_csv_data, write_csv


def add_missing_ids(input_dir: str, output_dir: str):
    pbar = tqdm(total=len(os.listdir(input_dir)))
    for filename in os.listdir(input_dir):
        data = get_csv_data(os.path.join(input_dir, filename))
        news_rows = []
        for row in data:
            if row['is_citing_in_meta']:
                news_rows.append({'id': row['citing'], 'title': '', 'author': '', 'editor': '', 'publisher': '', 'volume': '', 'venue': '', 'issue': '', 'pub_date': '', 'page': '', 'type': ''})
            elif row['is_cited_in_meta']:
                news_rows.append({'id': row['cited'], 'title': '', 'author': '', 'editor': '', 'publisher': '', 'volume': '', 'venue': '', 'issue': '', 'pub_date': '', 'page': '', 'type': ''})
        output_filepath = os.path.join(output_dir, filename)
        write_csv(output_filepath, news_rows)
        pbar.update()
    pbar.close()
            

if __name__ == '__main__': # pragma: no cover
    arg_parser = ArgumentParser('add_missing_ids.py', description='This script outputs Mtea input CSV files starting from CSVs containing missing identifiers in Meta')
    arg_parser.add_argument('-c', '--csv', dest='input_dir', required=True,
                            help='The directory containing the input CSV files')
    arg_parser.add_argument('-o', '--output', dest='output_dir', required=True,
                            help='The output directory')
    args = arg_parser.parse_args()
    add_missing_ids(args.input_dir, args.output_dir)


#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (c) 2022 Arcangelo Massari <arcangelo.massari@unibo.it>
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.


from __future__ import annotations

import csv
import json
import os
import sys
from contextlib import contextmanager
from datetime import datetime
from pathlib import Path
from time import sleep
from typing import Dict, List, Set
from zipfile import ZIP_DEFLATED, ZipFile

from _collections_abc import dict_keys
from bs4 import BeautifulSoup
from requests import ReadTimeout, get
from requests.exceptions import ConnectionError

from oc_meta.lib.cleaner import Cleaner


def get_csv_data(filepath:str) -> List[Dict[str, str]]:
    if not os.path.splitext(filepath)[1].endswith('.csv'):
        return list()
    field_size_changed = False
    cur_field_size = 128
    data = list()
    while not data:
        try:
            with open(filepath, 'r', encoding='utf8') as data_initial:
                valid_data = (Cleaner(line.replace('\0','')).normalize_spaces() for line in data_initial)
                data = list(csv.DictReader(valid_data, delimiter=','))
        except csv.Error:
            cur_field_size *= 2
            csv.field_size_limit(cur_field_size)
            field_size_changed = True
    if field_size_changed:
        csv.field_size_limit(128)
    return data

def pathoo(path):
    if not os.path.isdir(os.path.dirname(path)):
        os.makedirs(os.path.dirname(path))

def write_csv(path:str, datalist:List[dict], fieldnames:list|dict_keys|None=None, method:str='w') -> None:
    if datalist:
        fieldnames = datalist[0].keys() if fieldnames is None else fieldnames
        pathoo(path)
        file_exists = os.path.isfile(path)
        with open(path, method, newline='', encoding='utf-8') as output_file:
            dict_writer = csv.DictWriter(f=output_file, fieldnames=fieldnames, delimiter=',', quotechar='"', quoting=csv.QUOTE_NONNUMERIC)
            if method == 'w' or (method == 'a' and not file_exists):
                dict_writer.writeheader()
            dict_writer.writerows(datalist)

def normalize_path(path:str) -> str:
    normal_path = path.replace('\\', '/').replace('/', os.sep)
    return normal_path

def init_cache(cache_filepath:str|None) -> Set[str]:
    completed = set()
    if cache_filepath:
        if not os.path.exists(cache_filepath):
            pathoo(cache_filepath)
        else:
            with open(cache_filepath, 'r', encoding='utf-8') as cache_file:
                completed = {line.rstrip('\n') for line in cache_file}
    return completed

@contextmanager
def suppress_stdout():
    with open(os.devnull, 'w') as devnull: #pragma: no cover
        old_stdout = sys.stdout
        sys.stdout = devnull
        try:  
            yield
        finally:
            sys.stdout = old_stdout

def sort_files(files_to_be_processed:list) -> list:
    if all(filename.replace('.csv', '').isdigit() for filename in files_to_be_processed):
        files_to_be_processed = sorted(files_to_be_processed, key=lambda filename: int(filename.replace('.csv', '')))
    elif all(filename.split('_')[-1].replace('.csv', '').isdigit() for filename in files_to_be_processed):
        files_to_be_processed = sorted(files_to_be_processed, key=lambda filename: int(filename.split('_')[-1].replace('.csv', '')))
    return files_to_be_processed

def zipdir(path, ziph):
    for root, _, files in os.walk(path):
        for file in files:
            ziph.write(os.path.join(root, file),
                       os.path.relpath(os.path.join(root, file),
                                       os.path.join(path, '..')))

def zipit(dir_list:list, zip_name:str) -> None:
    zipf = ZipFile(file=zip_name, mode='w', compression=ZIP_DEFLATED, allowZip64=True)
    for dir in dir_list:
        zipdir(dir, zipf)
    zipf.close()

def zip_files_in_dir(src_dir:str, dst_dir:str, replace_files:bool=False) -> None:
    '''
    This method zips files individually in all directories starting from a specified root directory. 
    In other words, this function does not zip the entire folder but individual files 
    while maintaining the folder hierarchy in the specified output directory.

    :params src_dir: the source directory
    :type src_dir: str
    :params dst_dir: the destination directory
    :type dst_dir: str
    :params replace_files: True if you want to replace the original unzipped files with their zipped versions. The dafult value is False
    :type replace_files: bool
    :returns: None
    '''
    for dirpath, _, filenames in os.walk(src_dir):
        for filename in filenames:
            src_path = os.path.join(dirpath, filename)
            dst_path = os.path.join(
                dst_dir, 
                str(Path(src_path).parent)
                    .replace(f'{src_dir}{os.sep}', ''))
            if not os.path.exists(dst_path):
                os.makedirs(dst_path)
            _, ext = os.path.splitext(filename)
            zip_path = os.path.join(dst_path, filename).replace(ext, '.zip')
            with ZipFile(file=zip_path, mode='w', compression=ZIP_DEFLATED, allowZip64=True) as zipf:
                zipf.write(src_path, arcname=filename)
            if replace_files:
                os.remove(src_path)

def unzip_files_in_dir(src_dir:str, dst_dir:str, replace_files:bool=False) -> None:
    '''
    This method unzips zipped files individually in all directories starting from a specified root directory. 
    In other words, this function does not unzip the entire folder but individual files 
    while maintaining the folder hierarchy in the specified output directory.

    :params src_dir: the source directory
    :type src_dir: str
    :params dst_dir: the destination directory
    :type dst_dir: str
    :params replace_files: True if you want to replace the original zipped files with their unzipped versions, defaults to [False]
    :type replace_files: bool
    :returns: None
    '''
    for dirpath, _, filenames in os.walk(src_dir):
        for filename in filenames:
            if os.path.splitext(filename)[1] == '.zip':
                src_path = os.path.join(dirpath, filename)
                dst_path = os.path.join(
                    dst_dir, 
                    str(Path(src_path).parent)
                        .replace(f'{src_dir}{os.sep}', ''))
                if not os.path.exists(dst_path):
                    os.makedirs(dst_path)
                with ZipFile(file=os.path.join(dst_path, filename), mode='r') as zipf:
                    zipf.extractall(dst_path)
                if replace_files:
                    os.remove(src_path)

def read_zipped_json(filepath:str) -> dict|None:
    '''
    This method reads a zipped json file.

    :params filepath: the zipped json file path
    :type src_dir: str
    :returns: dict -- It returns the json file as a dictionary
    '''
    with ZipFile(filepath, 'r') as zipf:
        for filename in zipf.namelist(): 
            with zipf.open(filename) as f:
                json_data = f.read()
                json_dict = json.loads(json_data.decode("utf-8"))
                return json_dict

def call_api(url:str, headers:str, r_format:str="json") -> dict|None:
    tentative = 3
    while tentative:
        tentative -= 1
        try:
            r = get(url, headers=headers, timeout=30)
            if r.status_code == 200:
                r.encoding = "utf-8"
                return json.loads(r.text) if r_format == "json" else BeautifulSoup(r.text, 'xml')
            elif r.status_code == 404:
                return None
        except ReadTimeout:
            # Do nothing, just try again
            pass
        except ConnectionError:
            # Sleep 5 seconds, then try again
            sleep(5)
    return None

def rm_tmp_csv_files(base_dir:str) -> None:
    for filename in os.listdir(base_dir):
        number = filename.split('_')[0]
        date = datetime.strptime(filename.split('_')[1].replace('.csv', ''), '%Y-%m-%dT%H-%M-%S')
        for other_filename in os.listdir(base_dir):
            other_number = other_filename.split('_')[0]
            other_date = datetime.strptime(other_filename.split('_')[1].replace('.csv', ''), '%Y-%m-%dT%H-%M-%S')
            if number == other_number and filename != other_filename:
                if date < other_date:
                    os.remove(os.path.join(base_dir, filename))
                elif other_date < date:
                    os.remove(os.path.join(base_dir, other_filename))


import argparse
import csv
import os
import re
import zipfile
from multiprocessing import Pool, cpu_count, Manager
from typing import List, Dict, Set
from rdflib import ConjunctiveGraph
from SPARQLWrapper import SPARQLWrapper, JSON
from tqdm import tqdm

def parse_identifiers(id_string: str) -> List[Dict[str, str]]:
    """
    Parse space-separated identifiers in the format schema:value
    Returns a list of dicts with 'schema' and 'value' keys
    """
    if not id_string or id_string.isspace():
        return []
        
    identifiers = []
    for identifier in id_string.strip().split():
        try:
            schema, value = identifier.split(':', 1)
            identifiers.append({
                'schema': schema.lower(),
                'value': value
            })
        except ValueError:
            continue
    return identifiers

def check_omids_existence(identifiers: List[Dict[str, str]], endpoint_url: str) -> Set[str]:
    """
    Query SPARQL endpoint to find OMIDs for a single identifier
    Returns set of found OMIDs
    """
    if not identifiers or len(identifiers) != 1:
        return set()
    
    identifier = identifiers[0]
    sparql = SPARQLWrapper(endpoint_url)
    
    query = f"""
    PREFIX datacite: <http://purl.org/spar/datacite/>
    PREFIX literal: <http://www.essepuntato.it/2010/06/literalreification/>
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    
    SELECT DISTINCT ?omid
    WHERE {{
        {{
            ?omid literal:hasLiteralValue "{identifier['value']}"^^xsd:string ;
                 datacite:usesIdentifierScheme datacite:{identifier['schema']} .
        }}
        UNION
        {{
            ?omid literal:hasLiteralValue "{identifier['value']}" ;
                 datacite:usesIdentifierScheme datacite:{identifier['schema']} .
        }}
    }}
    """
    
    sparql.setQuery(query)
    sparql.setReturnFormat(JSON)
    
    try:
        results = sparql.query().convert()
        omids = set()
        for result in results["results"]["bindings"]:
            omid = result["omid"]["value"]
            omids.add(omid)
        return omids
    except Exception as e:
        print(f"SPARQL query failed for identifier {identifier}: {str(e)}")
        return set()

def find_file(rdf_dir: str, dir_split_number: int, items_per_file: int, uri: str, zip_output_rdf: bool) -> str|None:
    """Find the ZIP file containing the entity data"""
    entity_regex: str = r'^(https:\/\/w3id\.org\/oc\/meta)\/([a-z][a-z])\/(0[1-9]+0)?([1-9][0-9]*)$'
    entity_match = re.match(entity_regex, uri)
    if entity_match:
        cur_number = int(entity_match.group(4))
        cur_file_split: int = 0
        while True:
            if cur_number > cur_file_split:
                cur_file_split += items_per_file
            else:
                break
        cur_split: int = 0
        while True:
            if cur_number > cur_split:
                cur_split += dir_split_number
            else:
                break
        short_name = entity_match.group(2)
        sub_folder = entity_match.group(3)
        cur_dir_path = os.path.join(rdf_dir, short_name, sub_folder, str(cur_split))
        extension = '.zip' if zip_output_rdf else '.json'
        cur_file_path = os.path.join(cur_dir_path, str(cur_file_split)) + extension
        return cur_file_path
    return None

def load_graph_from_zip(zip_path: str) -> ConjunctiveGraph|None:
    """Load RDF graph from a ZIP file"""
    try:
        graph = ConjunctiveGraph()
        with zipfile.ZipFile(zip_path, 'r') as z:
            json_files = [f for f in z.namelist() if f.endswith('.json')]
            if not json_files:
                return None
            with z.open(json_files[0]) as f:
                graph.parse(data=f.read(), format='json-ld')
        return graph
    except Exception as e:
        print(f"Error loading graph from {zip_path}: {str(e)}")
        return None

def process_csv_file(args: tuple) -> Dict:
    """
    Process a single CSV file and check its identifiers
    Returns statistics about processed rows and found/missing OMIDs
    """
    file_path, endpoint_url, root_dir = args
    stats = {
        'total_rows': 0,
        'rows_with_ids': 0,
        'total_identifiers': 0,
        'identifiers_with_omids': 0,
        'identifiers_without_omids': 0,
        'identifiers_details': [],  # To store detailed info about each identifier
        'graphs_found': 0,
        'graphs_missing': 0,
        'processed_omids': {}  # Track which OMID was found and in which row
    }
    
    try:
        with open(file_path, 'r', encoding='utf-8') as f:
            reader = csv.DictReader(f)
            for row in reader:
                stats['total_rows'] += 1
                
                id_column = row.get('id', '').strip()
                if not id_column:
                    continue
                    
                stats['rows_with_ids'] += 1
                identifiers = parse_identifiers(id_column)
                stats['total_identifiers'] += len(identifiers)
                
                # Check each identifier individually
                for identifier in identifiers:
                    omids = check_omids_existence([identifier], endpoint_url)
                    has_omid = len(omids) > 0
                    
                    if has_omid:
                        if len(omids) > 1:
                            print(f"Warning: Identifier {identifier['schema']}:{identifier['value']} "
                                  f"is associated with multiple OMIDs: {omids}")
                        stats['identifiers_with_omids'] += 1
                        # Try to load the graph for each OMID
                        for omid in omids:
                            zip_path = find_file(root_dir, 10000, 1000, omid, True)
                            if zip_path and os.path.exists(zip_path):
                                graph = load_graph_from_zip(zip_path)
                                if graph is not None:
                                    stats['graphs_found'] += 1
                                    stats['processed_omids'][omid] = {
                                        'row': stats['total_rows'],
                                        'identifier': f"{identifier['schema']}:{identifier['value']}",
                                        'file': file_path
                                    }
                                else:
                                    stats['graphs_missing'] += 1
                            else:
                                stats['graphs_missing'] += 1
                    else:
                        stats['identifiers_without_omids'] += 1
                    
                    stats['identifiers_details'].append({
                        'schema': identifier['schema'],
                        'value': identifier['value'],
                        'has_omid': has_omid,
                        'row_number': stats['total_rows']
                    })
                
    except Exception as e:
        print(f"Error processing {file_path}: {str(e)}")
        
    return stats

def main():
    parser = argparse.ArgumentParser(description="Check MetaProcess results by verifying input CSV identifiers")
    parser.add_argument("directory", help="Directory containing input CSV files")
    parser.add_argument("--root", required=True,
                       help="Root directory containing the ZIP files with JSON-LD data")
    parser.add_argument("--endpoint", required=True, help="SPARQL endpoint URL")
    parser.add_argument("--show-missing", action="store_true", 
                       help="Show details of identifiers without associated OMIDs")
    args = parser.parse_args()
    
    # Find all CSV files in the directory and subdirectories
    csv_files = []
    for root, _, files in os.walk(args.directory):
        csv_files.extend(
            os.path.join(root, f) for f in files if f.endswith('.csv')
        )
    
    if not csv_files:
        print(f"No CSV files found in {args.directory}")
        return
        
    print(f"Found {len(csv_files)} CSV files to process")
    
    # Process files in parallel (senza shared data)
    with Pool(cpu_count()) as pool:
        process_args = [(f, args.endpoint, args.root) for f in csv_files]
        results = list(tqdm(
            pool.imap(process_csv_file, process_args),
            total=len(csv_files),
            desc="Processing CSV files"
        ))
    
    # Aggregate results
    total_rows = sum(r['total_rows'] for r in results)
    total_rows_with_ids = sum(r['rows_with_ids'] for r in results)
    total_identifiers = sum(r['total_identifiers'] for r in results)
    total_with_omids = sum(r['identifiers_with_omids'] for r in results)
    total_without_omids = sum(r['identifiers_without_omids'] for r in results)
    total_graphs_found = sum(r.get('graphs_found', 0) for r in results)
    total_graphs_missing = sum(r.get('graphs_missing', 0) for r in results)
    
    # Collect all OMID occurrences
    all_omids = {}
    for result in results:
        for omid, details in result['processed_omids'].items():
            if omid not in all_omids:
                all_omids[omid] = []
            all_omids[omid].append(details)
    
    # Find duplicates
    duplicates = {omid: occurrences for omid, occurrences in all_omids.items() if len(occurrences) > 1}
    
    # Print summary
    print("\nResults Summary:")
    print(f"Total rows processed: {total_rows}")
    print(f"Rows containing identifiers: {total_rows_with_ids}")
    print(f"Total identifiers found: {total_identifiers}")
    print(f"Identifiers with associated OMIDs: {total_with_omids} ({(total_with_omids/total_identifiers*100):.2f}%)")
    print(f"Identifiers without OMIDs: {total_without_omids} ({(total_without_omids/total_identifiers*100):.2f}%)")
    print(f"Graphs found in ZIP files: {total_graphs_found}")
    print(f"Graphs missing from ZIP files: {total_graphs_missing}")
    print(f"Unique OMIDs found: {len(all_omids)}")
    
    if duplicates:
        print("\nOMIDs found multiple times:")
        for omid, occurrences in duplicates.items():
            print(f"\nOMID {omid} appeared {len(occurrences)} times:")
            for occurrence in occurrences:
                print(f"  Row {occurrence['row']} in {occurrence['file']}")
                print(f"  Identifier: {occurrence['identifier']}")

if __name__ == "__main__":
    main() 


from __future__ import annotations

import os
from argparse import ArgumentParser

from fakeredis import FakeStrictRedis
from oc_ds_converter.datasource.redis import RedisDataSource
from tqdm import tqdm

from oc_meta.lib.file_manager import get_csv_data, write_csv


def discard_existing_res(csv_filepath: str, output_filepath: str, r: FakeStrictRedis|RedisDataSource = FakeStrictRedis()):
    r = r if isinstance(r, FakeStrictRedis) else r._r
    pbar = tqdm(total=len(os.listdir(csv_filepath)))
    if not os.path.exists(output_filepath):
        os.makedirs(output_filepath)
    counter = 1
    for filename in os.listdir(csv_filepath):
        data = get_csv_data(os.path.join(csv_filepath, filename))
        all_identifiers = list()
        preexisting_identifiers = set()
        for i, row in enumerate(data):
            all_identifiers.extend(row['id'].split())
            if i % 50000 == 0:
                identifiers_in_r = r.mget(all_identifiers)
                preexisting_identifiers.update({identifier for i, identifier in enumerate(all_identifiers) if identifiers_in_r[i]})
                all_identifiers = list()
        identifiers_in_r = r.mget(all_identifiers)
        preexisting_identifiers.update({identifier for i, identifier in enumerate(all_identifiers) if identifiers_in_r[i]})
        output_data = list()
        for row in data:
            if not all(identifier in preexisting_identifiers for identifier in row['id'].split()):
                output_data.append(row)
        if output_data:
            cur_output_filepath = os.path.join(output_filepath, f'{str(counter)}.csv')
            write_csv(cur_output_filepath, output_data)
            counter += 1
        pbar.update()
    pbar.close()


if __name__ == '__main__': # pragma: no cover
    arg_parser = ArgumentParser('discard_existing_res.py', description='This script removes rows from metadata CSV tables if such rows refer to bibliographic resources that are in OpenCitations Meta')
    arg_parser.add_argument('-c', '--csv', dest='csv_filepath', required=True,
                            help='The directory containing the metadata CSV files')
    arg_parser.add_argument('-o', '--output', dest='output_filepath', required=True,
                                help='The directory to store the output CSV files')
    arg_parser.add_argument('-t', '--testing', dest='testing', required=False, action='store_true',
                                help='Use fakeredis')
    args = arg_parser.parse_args()
    r = FakeStrictRedis() if args.testing else RedisDataSource('DB-META-BR')
    discard_existing_res(args.csv_filepath, args.output_filepath, r)


import argparse
import yaml
import zenodopy

def load_config(yaml_file_path):
    with open(yaml_file_path, 'r') as stream:
        try:
            return yaml.safe_load(stream)
        except yaml.YAMLError as exc:
            print(exc)
            return None

def main(config_file):
    # Load configuration
    config = load_config(config_file)
    if config is None:
        print("Failed to load the configuration.")
        return

    # Determine if we are using the sandbox environment
    sandbox = "sandbox" in config['zenodo_url']

    # Initialize zenodopy client
    zeno = zenodopy.Client(token=config['access_token'], sandbox=sandbox)

    project_id = config.get('project_id')
    if project_id:
        # Set the existing project
        zeno.set_project(project_id)
    else:
        # Create a new project (deposition)
        title = config.get('title', '')
        description = config.get('description', '')
        upload_type = config.get('upload_type', '')
        zeno.create_project(title=title, description=description, upload_type=upload_type)

    # Upload files to the project
    for file_path in config['files']:
        zeno.upload_file(file_path)

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Upload files to Zenodo using a YAML configuration.")
    parser.add_argument('config_file', type=str, help="Path to the YAML configuration file.")
    args = parser.parse_args()

    main(args.config_file)